// $Id: autoaction.cc,v 1.6 2003/03/15 04:47:06 toast Exp $

// Std C++ Headers
#include <string>

// OOMon Headers
#include "autoaction.h"
#include "irc.h"
#include "util.h"
#include "vars.h"
#include "main.h"
#include "botsock.h"


static void
doKill(const std::string & nick, const std::string & reason)
{
  if (vars[VAR_AUTO_PILOT]->getBool())
  {
    server.kill("Auto-Kill", nick, reason);
  }
  else
  {
    ::SendAll(".kill " + nick + " " + reason, UF_OPER);
  }
}


static void
doKline(const std::string & mask, int duration, const std::string & reason)
{
  if (vars[VAR_AUTO_PILOT]->getBool())
  {
    server.kline("Auto-Kline", duration, mask, reason);
  }
  else
  {
    if (duration > 0)
    {
      ::SendAll(".kline " + IntToStr(duration) + " " + mask + " " + reason,
	UF_OPER | UF_KLINE);
    }
    else
    {
      ::SendAll(".kline " + mask + " " + reason, UF_OPER | UF_KLINE);
    }
  }
}


static void
doDline(const std::string & mask, const std::string & reason)
{
  if (vars[VAR_AUTO_PILOT]->getBool())
  {
    server.dline("Auto-Dline", mask, reason);
  }
  else
  {
    ::SendAll(".dline " + mask + " " + reason, UF_OPER | UF_DLINE);
  }
}


void
doAction(const std::string & nick, const std::string & userhost,
  const BotSock::Address & ip, const AutoAction & action, int duration,
  const std::string & reason, bool suggestKlineAfterKill)
{
  std::string::size_type at = userhost.find('@');
  if (std::string::npos == at)
  {
    // user@host is invalid format! just give up.
    return;
  }

  std::string user = userhost.substr(0, at);
  std::string host = userhost.substr(at + 1);
  std::string domain = getDomain(host, true);

  if ((user.length() > 1) && (user[0] == '~'))
  {
    user = user.substr(1);
  }
  else if (user.length() > 8)
  {
    user = user.substr(user.length() - 8);
  }

  if ((domain.length() > 1) && (domain[domain.length() - 1] == '.'))
  {
    domain = domain + "*";
  }
  else
  {
    domain = "*" + domain;
  }

  switch (action)
  {
  case ACTION_KILL:
    doKill(nick, reason);
    if (suggestKlineAfterKill && vars[VAR_AUTO_PILOT]->getBool())
    {
      std::string mask("*" + user + "@" + domain);
      ::SendAll(".kline " + mask + " " + reason, UF_OPER | UF_KLINE);
    }
    break;

  case ACTION_KLINE:
    doKline("*" + user + "@" + domain, duration, reason);
    break;

  case ACTION_KLINE_HOST:
    doKline("*@" + host, duration, reason);
    break;

  case ACTION_KLINE_DOMAIN:
    doKline("*@" + domain, duration, reason);
    break;

  case ACTION_KLINE_IP:
    if (ip != INADDR_NONE)
    {
      doKline("*@" + BotSock::inet_ntoa(ip), duration, reason);
    }
    break;

  case ACTION_KLINE_USERNET:
    if (ip != INADDR_NONE)
    {
      doKline("*" + user + "@" + classCMask(BotSock::inet_ntoa(ip)), duration,
	reason);
    }
    break;

  case ACTION_KLINE_NET:
    if (ip != INADDR_NONE)
    {
      doKline("*@" + classCMask(BotSock::inet_ntoa(ip)), duration, reason);
    }
    break;

  case ACTION_DLINE_IP:
    if ((ip != INADDR_NONE) && (ip != INADDR_ANY))
    {
      doDline(BotSock::inet_ntoa(ip), reason);
    }
    break;

  case ACTION_DLINE_NET:
    if ((ip != INADDR_NONE) && (ip != INADDR_ANY))
    {
      doDline(classCMask(BotSock::inet_ntoa(ip)), reason);
    }
    break;
  }
}

