// $Id: botexcept.h,v 1.5 2003/05/25 17:17:50 toast Exp $

#ifndef __BOTEXCEPT_H__
#define __BOTEXCEPT_H__

// Std C++ Headers
#include <string>
#include <stdexcept>

// Std C Headers
#include <errno.h>

// OOMon Headers
#include "botdb.h"


namespace OOMon
{

  class oomon_error
  {
  public:
    oomon_error(const std::string & arg) { _what = arg; };
    virtual std::string what() const { return _what; };

  private:
    std::string _what;
  };

  class errno_error : public oomon_error
  {
  public:
    errno_error(const std::string & arg) : oomon_error(arg)
      { this->_errno = errno; };
    std::string why() const { return strerror(this->_errno); };

  private:
    int _errno;
  };

  class botsock_error : public oomon_error {
  public:
    botsock_error(const string & arg): oomon_error(arg) { };
  };

  class open_error : public errno_error {
  public:
    open_error(const string & arg): errno_error(arg) { };
  };

  class close_error : public errno_error {
  public:
    close_error(const string & arg): errno_error(arg) { };
  };

  class socket_error : public errno_error {
  public:
    socket_error(const string & arg): errno_error(arg) { };
  };

  class accept_error : public errno_error {
  public:
    accept_error(const string & arg): errno_error(arg) { };
  };

  class timeout_error : public botsock_error {
  public:
    timeout_error(const string & arg): botsock_error(arg) { };
  };

  class ready_for_accept : public botsock_error {
  public:
    ready_for_accept(const string & arg): botsock_error(arg) { };
  };

  class botdb_error : public oomon_error {
  public:
    botdb_error(const string & arg): oomon_error(arg)
      { this->_errno = BotDB::db_errno(); };
    std::string why() const { return BotDB::strerror(this->_errno); };

  private:
    int _errno;
  };

  class norecord_error : public oomon_error {
  public:
    norecord_error(const string & arg): oomon_error(arg) { };
  };

  class watch_error : public oomon_error {
  public:
    watch_error(const string & arg): oomon_error(arg) { };
  };

  class invalid_watch_name : public watch_error {
  public:
    invalid_watch_name(const string & arg): watch_error(arg) { };
  };

  class invalid_watch_value : public watch_error {
  public:
    invalid_watch_value(const string & arg): watch_error(arg) { };
  };

  class invalid_action : public oomon_error {
  public:
    invalid_action(const string & arg): oomon_error(arg) { };
  };

};

#endif /* __BOTEXCEPT_H__ */

