// $Id: botsock.h,v 1.13 2003/02/23 23:35:51 toast Exp $

#ifndef __BOTSOCK_H__
#define __BOTSOCK_H__

// Std C++ headers
#include <string>

// Std C headers
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <time.h>

// OOMon C++ headers
#include "oomon.h"


#ifndef INADDR_NONE
# define INADDR_NONE	0xffffffff
#endif
#ifndef INADDR_ANY
# define INADDR_ANY	0
#endif


class BotSock
{
public:
  typedef in_addr_t Address;
  typedef in_port_t Port;

  BotSock(const bool blocking = false, const bool lineBuffered = false);
  BotSock(const BotSock *listener, const bool blocking = false,
    const bool lineBuffered = false);
  virtual ~BotSock();

  bool connect(const Address address, const Port port);
  bool connect(const std::string & address, const Port port);

  bool reset();

  bool listen(const Port port = 0, const int backlog = 1);

  int read(void *buffer, int size);
  int write(void *buffer, int size);
  int write(const std::string & text);

  bool process(const fd_set &, const fd_set &);
  void setFD(fd_set &, fd_set &) const;

  void bindTo(const std::string & address);

  Address getLocalAddress();
  Port getLocalPort();

  Address getRemoteAddress();
  Port getRemotePort();

  bool setBlocking(const bool value);
  bool isBlocking() const { return this->blocking; };

  void setBuffering(const bool value);
  bool isBuffering() const { return this->lineBuffered; };

  bool isConnected() const { return this->connected; };
  bool isConnecting() const { return this->connecting; };
  bool isListening() const { return this->listening; };

  void setTimeout(const time_t value) { this->timeout = value; };
  time_t getTimeout() const { return this->timeout; };
  time_t getIdle() const { return (time(NULL) - this->lastActivity); };
  void gotActivity() { this->lastActivity = time(NULL); };

  static Address nsLookup(const std::string & address);
  static std::string inet_ntoa(const BotSock::Address & address);
  static Address inet_addr(const std::string & address)
    { return ::inet_addr(address.c_str()); };

  static const BotSock::Address vhost_netmask;

  virtual bool onConnect() { return true; };

protected:
  virtual bool onRead(std::string) { return true; };

private:
  void setOptions();
  bool bind(const BotSock::Address & address = INADDR_ANY,
    const BotSock::Port & port = 0);

  std::string buffer;
  Address bindAddress;
  time_t timeout, lastActivity;
  bool connected, connecting, listening, blocking, lineBuffered;
  int plug, backlog;
};


#endif /* __BOTSOCK_H__ */

