// $Id: config.cc,v 1.27 2003/05/17 18:52:26 toast Exp $

// C headers
#include <stdlib.h>
#include <unistd.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <netdb.h>
#include <arpa/inet.h>

#include "oomon.h"

// C++ headers
#include <iostream>
#include <fstream>
#include <string>
#include <vector>
#include "config.h"
#include "util.h"
#include "help.h"
#include "trap.h"
#include "vars.h"


#ifdef DEBUG
# define CONFIG_DEBUG
#endif


#ifndef MAXHOSTNAMELEN
# define MAXHOSTNAMELEN 256
#endif


OperDef::OperDef()
{
  Handle = Mask = Passwd = "";
  Flags = UF_NONE;
}


OperDef::OperDef(const OperDef & def)
{
  Handle = def.Handle;
  Mask = def.Mask;
  Passwd = def.Passwd;
  Flags = def.Flags;
}


LinkDef::LinkDef()
{
  Handle = Mask = Passwd = "";
  Flags = LF_NONE;
}


LinkDef::LinkDef(const LinkDef & def)
{
  Handle = def.Handle;
  Mask = def.Mask;
  Passwd = def.Passwd;
  Flags = def.Flags;
}


ConnDef::ConnDef()
{
  Handle = Hostname = Passwd = "";
  Flags = CF_NONE;
  port = 0;
}


ConnDef::ConnDef(const ConnDef & def)
{
  Handle = def.Handle;
  Hostname = def.Hostname;
  Passwd = def.Passwd;
  Flags = def.Flags;
  port = def.port;
}


YLine::YLine()
{
  YClass = "";
  Description = "";
}


YLine::YLine(const YLine & def)
{
  YClass = def.YClass;
  Description = def.Description;
}

OperList Config::Opers;
BotSock::Port Config::port = 4000;
StrList Config::Exceptions;
StrList Config::Spoofers;
LinkList Config::Links;
ConnList Config::Connections;
YLineList Config::YLines;
std::string Config::Nick, Config::UserName, Config::IRCName, Config::OperNick, Config::OperPass;
std::string Config::OurHostName;
std::string Config::LogFile = "oomon.log";
std::string Config::MOTD = "motd.txt";
struct server Config::Server;
std::string Config::proxyVhost;
std::string Config::helpFilename = DEFAULT_HELPFILE;
std::string Config::settingsFile = DEFAULT_SETTINGSFILE;
std::string Config::userDBFile = DEFAULT_USERDBFILE;

// AddOper(Handle, Mask, Passwd, Flags)
//
// This does the actually adding of an Oper to the list.
// The user@host mask is the only required parameter.
// Passwords and handles are highly recommended. In fact, a handle is
// required for authorization if you have a password! :P
// Supported flags are: (case sensitive!)
//  C - user can request channel ops
//  D - user can dline
//  G - user can gline
//  K - user can kline, kill, etc.
//  M - user is a master and can control linking
//  N - user can see nick changes
//  O - user is an oper (pretty mandatory)
//  R - user can remotely control other bots
//  W - user can see WALLOPS
//  X - user can see client connects/disconnects
//
void Config::AddOper(const std::string & Handle, const std::string & Mask,
	const std::string & Passwd, const std::string & Flags)
{
  if (std::string::npos != Mask.find('@'))
  {
    OperDef temp;
    temp.Handle = Handle;
    temp.Mask = Mask;
    temp.Passwd = Passwd;
    temp.Flags = UF_NONE;
    if (Flags.find('C') != std::string::npos)
      temp.Flags |= UF_CHANOP;
    if (Flags.find('D') != std::string::npos)
      temp.Flags |= UF_DLINE;
    if (Flags.find('G') != std::string::npos)
      temp.Flags |= UF_GLINE;
    if (Flags.find('K') != std::string::npos)
      temp.Flags |= UF_KLINE;
    if (Flags.find('M') != std::string::npos)
      temp.Flags |= UF_MASTER;
    if (Flags.find('N') != std::string::npos)
      temp.Flags |= UF_NICK;
    if (Flags.find('O') != std::string::npos)
      temp.Flags |= UF_OPER;
    if (Flags.find('R') != std::string::npos)
      temp.Flags |= UF_REMOTE;
    if (Flags.find('W') != std::string::npos)
      temp.Flags |= UF_WALLOPS;
    if (Flags.find('X') != std::string::npos)
      temp.Flags |= UF_CONN;
    Opers.push_back(temp);
    #ifdef CONFIG_DEBUG
      cout << "Oper: " << Handle << ", " << Mask << ", " << Passwd << ", " << Flags << endl;
    #endif
  }
}


// AddException(Mask)
//
// Adds an exception mask to the list
//
void Config::AddException(const std::string & Mask)
{
  if (Mask != "") {
    Exceptions.push_back(Mask);
    #ifdef CONFIG_DEBUG
      cout << "Exception: " << Mask << endl;
    #endif
  }
}


// AddSpoofer(Mask)
//
// Adds an spoofer ip mask to the list
//
void Config::AddSpoofer(const std::string & Mask)
{
  if (Mask != "") {
    Spoofers.push_back(Mask);
    #ifdef CONFIG_DEBUG
      cout << "Spoofer: " << Mask << endl;
    #endif
  }
}


// AddLink(Handle, Mask, Passwd, Flags)
//
// Adds a link definition to the list.
// A handle and host mask are required! Passwords are highly recommended!
// Supported flags are: (case sensitive!)
//
//
void Config::AddLink(const std::string & Handle, const std::string & Mask,
	const std::string & Passwd, const std::string & Flags)
{
  if ((Handle != "") && (Mask != "")) {
    LinkDef temp;
    temp.Handle = Handle;
    temp.Mask = Mask;
    temp.Passwd = Passwd;
    //temp.Flags = Flags;
    Links.push_back(temp);
    #ifdef CONFIG_DEBUG
      cout << "Link: " << Handle << ", " << Mask << ", " << Passwd << ", " << Flags << endl;
    #endif
  }
}


// AddConn(Handle, Hostname, Passwd, Flags, port)
//
// Adds a connection definition to the list
// A handle, hostname, and port number are required! Password requirements
// depend on the bot you are connecting to.
// Supported flags are the same as for AddLink()
//
void Config::AddConn(const std::string & Handle, const std::string & Hostname,
  const std::string & Passwd, const std::string & Flags,
  const BotSock::Port port)
{
  if ((Handle != "") && (Hostname != "") && (port != 0)) {
    ConnDef temp;
    temp.Handle = Handle;
    temp.Hostname = Hostname;
    temp.Passwd = Passwd;
    //temp.Flags = Flags;
    temp.port = port;
    Connections.push_back(temp);
    #ifdef CONFIG_DEBUG
      cout << "Conn: " << Handle << ", " << Hostname << ", " << Passwd << ", " << Flags << ", " << port << endl;
    #endif
  }
}


// AddYLine(YClass, Description)
//
// Adds a connection class (Y-line) to the list
// A YClass and Description are required.
//
void Config::AddYLine(const std::string & YClass,
  const std::string & Description)
{
  if (Description != "")
  {
    YLine temp;
    temp.YClass = YClass;
    temp.Description = Description;
    YLines.push_back(temp);
    #ifdef CONFIG_DEBUG
      cout << "Y-Line: " << YClass << ", " << Description << endl;
    #endif
  }
}


// Load(filename)
//
// This does the actual loading of the configuration file.
//
// I ought to trim any leading white space before parsing, but oh well.
//
void Config::Load(const std::string filename)
{
  char		line[MAX_BUFF];
  char		*key,
		*value;
  std::string	handle,
		mask,
		hostname,
		passwd,
		flags,
		yclass;
  int		port;
  StrVector Parameters;
 
  #ifdef CONFIG_DEBUG
    cout << "Reading configuration file (" << filename << ")" << endl;
  #endif

  Help::flush();

  ifstream CfgFile(filename.c_str());
  while (CfgFile.getline(line, MAX_BUFF-1))
  {
    if (line[0] == '#')
    {
      continue;
    }

    key = strtok(line, ":");
    if (key == (char *) NULL)
    {
      continue;
    }
    value = strtok((char *) NULL, "\r\n");
    if (value == (char *) NULL)
    {
      continue;
    }
      
    StrSplit(Parameters, value, ":");

    switch (*key)
    {
      case 'b':
      case 'B':
	if (Parameters.size() > 5)
	{
	  // Bot parameters
	  #ifdef CONFIG_DEBUG
	    cout << "Bot: ";
	    //Parameters.Print();
	  #endif
	  Nick = Parameters[0];
	  UserName = Parameters[1];
	  OurHostName = Parameters[2];
	  IRCName = Parameters[3];
	  OperNick = Parameters[4];
	  OperPass = Parameters[5];
	}
	break;
      case 'c':
      case 'C':
	if (Parameters.size() > 4)
	{
	  // Connection parameters
	  handle = Parameters[0];
	  hostname = Parameters[1];
	  passwd = Parameters[2];
	  flags = Parameters[3];
	  port = atoi(Parameters[4].c_str());
	  AddConn(handle, hostname, passwd, UpCase(flags), port);
	}
	break;
      case 'e':
      case 'E':
	if (Parameters.size() > 0)
	{
	  // Exception
	  AddException(Parameters[0]);
	}
	break;
      case 'f':
      case 'F':
	if (Parameters.size() > 0)
	{
	  // Spoofers
	  AddSpoofer(Parameters[0]);
	}
	break;
      case 'g':
      case 'G':
	if (Parameters.size() > 0)
	{
	  // Log file
	  LogFile = Parameters[0];
        }
	break;
      case 'h':
      case 'H':
	if (Parameters.size() > 0)
	{
	  // Help file
	  helpFilename = Parameters[0];
	}
	break;
      case 'i':
      case 'I':
	if (Parameters.size() > 0)
	{
	  // Include file
	  Load(Parameters[0]);
	}
	break;
      case 'l':
      case 'L':
	if (Parameters.size() > 3)
	{
	  // Link parameters
	  handle = Parameters[0];
	  mask = Parameters[1];
	  passwd = Parameters[2];
	  flags = Parameters[3];
	  AddLink(handle, mask, passwd, UpCase(flags));
	}
	break;
      case 'm':
      case 'M':
	if (Parameters.size() > 0)
	{
	  // MOTD file
	  MOTD = Parameters[0];
	}
	break;
      case 'o':
      case 'O':
	if (Parameters.size() > 3)
	{
	  // Oper definition
	  handle = Parameters[0];
	  mask = Parameters[1];
	  passwd = Parameters[2];
	  flags = Parameters[3];
	  AddOper(handle, mask, passwd, UpCase(flags));
	}
	break;
      case 'p':
      case 'P':
	if (Parameters.size() > 0)
	{
	  // Port number
	  port = atoi(Parameters[0].c_str());
	}
	break;
      case 's':
      case 'S':
	if (Parameters.size() > 3)
	{
	  // Server Parameters
	  #ifdef CONFIG_DEBUG
	    cout << "Server: ";
	    //Parameters.Print();
	  #endif
	  Server.Hostname = Parameters[0];
	  Server.port = atoi(Parameters[1].c_str());
	  Server.Password = Parameters[2];
	  Server.Channels =  Parameters[3];
	}
	break;
      case 't':
      case 'T':
	if (Parameters.size() > 0)
	{
	  // OOMon Settings file
	  settingsFile = Parameters[0];
	}
	break;
      case 'u':
      case 'U':
	if (Parameters.size() > 0)
	{
	  // User settings file
	  userDBFile = Parameters[0];
	}
	break;
      case 'w':
      case 'W':
	if (Parameters.size() > 0)
	{
	  // WinGate/Proxy checker virtual hostname
	  proxyVhost = Parameters[0];
	}
	break;
      case 'y':
      case 'Y':
	if (Parameters.size() > 1)
	{
	  // User classes (Y-lines)
	  #ifdef CONFIG_DEBUG
	    cout << "Y-line: ";
	    //Parameters.Print();
	  #endif
	  yclass = Parameters[0];
	  hostname = Parameters[1];
	  AddYLine(yclass, hostname);
	}
        break;
    }
  }
  CfgFile.close();

  #ifdef CONFIG_DEBUG
    cout << "Completed reading configuration (" << filename << ")" << endl;
  #endif
}


// Clear()
//
// Clears the current configuration. This is usually done immediately before
// loading the configuration from file.
//
void Config::Clear()
{
  Opers.clear();
  Exceptions.clear();
  Spoofers.clear();
  Links.clear();
  Connections.clear();
  YLines.clear();
  port = 4000;
  Server.Hostname = "";
  Server.port = 0;
  Server.Password = "";
  Server.Channels = "";
  LogFile = "oomon.log";
  MOTD = "motd.txt";
}


// Auth(Handle, UserHost, Password, Flags, NewHandle)
//
// Attempts to receive authorization.
//
// Handle and Password should be entered by the user.  The UserHost is taken
// from the user's address.  Flags and NewHandle are returned to the caller
// as defined in the configuration file.
//
// Returns true if authorization is granted, false otherwise.
//
// Note that Flags and NewHandle are NOT modified if authorization fails.
//
bool Config::Auth(const std::string & Handle, const std::string & UserHost,
	const std::string & Password, int & Flags, std::string & NewHandle)
{
#ifdef CONFIG_DEBUG
  cout << "Config::Auth(Handle = \"" << Handle << "\", UserHost = \"" <<
    UserHost << "\", Password = \"" << Password << "\", ...)" << endl;
#endif
  for (OperList::iterator pos = Opers.begin(); pos != Opers.end(); ++pos)
  {
    if (Same(pos->Handle, Handle) && MatchesMask(UserHost, pos->Mask))
    {
#ifdef CONFIG_DEBUG
      cout << "Config::Auth(): found match" << endl;
#endif
      if (ChkPass(pos->Passwd, Password))
      {
#ifdef CONFIG_DEBUG
        cout << "Config::Auth(): password is correct" << endl;
#endif
        if (pos->Handle != "")
	{
	  // Only people with registered handles should be allowed to do
	  // anything
          NewHandle = pos->Handle;
          Flags = pos->Flags | UF_AUTHED;
        }
	else
	{
	  // You might want to set up an open O: line *@* in the config file
	  // (with no nick) to allow anyone to chat
	  Flags = UF_AUTHED;
        }
        return true;
      }
      else
      {
#ifdef CONFIG_DEBUG
        cout << "Config::Auth(): incorrect password" << endl;
#endif
      }
    }
  }
  return false;
}

std::string Config::GetHostName()
{
  return OurHostName;
}

std::string Config::GetUserName()
{
  if (UserName == "")
  {
    char *login = getlogin();

    if (login && *login)
      return login;
    else
      return "oomon";
  }
  else
  {
    return UserName;
  }
}

bool Config::IsOKHost(const std::string & userhost)
{
  for (StrList::iterator pos = Exceptions.begin();
    pos != Exceptions.end(); ++pos)
  {
    if (MatchesMask(userhost, *pos))
    {
      return true;
    }
  }
  return false;
}

bool Config::IsOKHost(const std::string & userhost, const std::string & ip)
{
  if (ip == "")
    return Config::IsOKHost(userhost);

  std::string user = userhost.substr(0, userhost.find('@'));
  std::string userip = user + "@" + ip;

  for (StrList::iterator pos = Exceptions.begin();
    pos != Exceptions.end(); ++pos)
  {
    if (MatchesMask(userhost, *pos) || MatchesMask(userip, *pos))
    {
      return true;
    }
  }
  return false;
}

bool Config::IsOper(const std::string & userhost)
{
  for (OperList::iterator pos = Opers.begin(); pos != Opers.end(); ++pos)
  {
    if (MatchesMask(userhost, pos->Mask))
    {
      return (pos->Flags & UF_OPER);
    }
  }
  return false;
}

bool Config::IsOper(const std::string & userhost, const std::string & ip)
{
  if (ip == "")
    return Config::IsOper(userhost);

  std::string user = userhost.substr(0, userhost.find('@'));
  std::string userip = user + "@" + ip;

  for (OperList::iterator pos = Opers.begin(); pos != Opers.end(); ++pos)
  {
    if (MatchesMask(userhost, pos->Mask) || MatchesMask(userip, pos->Mask))
    {
      return (pos->Flags & UF_OPER);
    }
  }
  return false;
}

bool Config::IsLinkable(const std::string & host)
{
  for (LinkList::iterator pos = Links.begin(); pos != Links.end(); ++pos)
  {
    if (MatchesMask(host, pos->Mask))
    {
      return true;
    }
  }
  return false;
}

bool Config::GetConn(std::string & BotHandle, std::string & Host,
  BotSock::Port & port, std::string & Passwd, int & Flags)
{
  for (ConnList::iterator pos = Connections.begin(); pos != Connections.end();
    ++pos)
  {
    if (Same(BotHandle, pos->Handle))
    {
      BotHandle = pos->Handle;
      Host = pos->Hostname;
      port = pos->port;
      Passwd = pos->Passwd;
      Flags = pos->Flags;
      return true;
    }
  }
  return false;
}

bool Config::AuthBot(const std::string & Handle, const std::string & Host,
	const std::string & Passwd)
{
  for (LinkList::iterator pos = Links.begin(); pos != Links.end(); ++pos)
  {
    if (Same(Handle, pos->Handle) && MatchesMask(Host, pos->Mask) &&
        (pos->Passwd == Passwd))
    {
      return true;
    }
  }
  return false;
}

std::string Config::GetYLineDescription(const std::string & YClass)
{
  for (YLineList::iterator pos = YLines.begin(); pos != YLines.end(); ++pos)
  {
    if (YClass == pos->YClass)
    {
      return pos->Description;
    }
  }
  return "";
}

std::string
Config::getProxyVhost()
{
  if (proxyVhost == "")
    return Config::GetHostName();
  else
    return proxyVhost;
}

bool Config::IsSpoofer(const std::string & ip)
{
  for (StrList::iterator pos = Spoofers.begin();
    pos != Spoofers.end(); ++pos)
  {
    if (MatchesMask(ip, *pos))
    {
      return true;
    }
  }
  return false;
}


bool
Config::saveSettings()
{
  ofstream file(settingsFile.c_str());

  if (file)
  {
    // Write header
    file << "# This file was generated by OOMon-" OOMON_VERSION << endl;
    file << "#" << endl;
    file << "# Please do not modify this file directly!" << endl;
    file << "#" << endl;

    // Save TRAP list
    TrapList::save(file);

    // Save SET variables
    vars.save(file);

    // Done!
    file.close();

    return true;
  }
  else
  {
    return false;
  }
}


bool
Config::loadSettings()
{
  ifstream file(settingsFile.c_str());

  if (file)
  {
    // First clear out the TRAP list!
    TrapList::clear();

    std::string line;

    while (std::getline(file, line))
    {
      std::string cmd = UpCase(FirstWord(line));

      if ((cmd.length() > 0) && (cmd[0] == '#'))
      {
	// Just ignore the comments
#ifdef CONFIG_DEBUG
        cout << "Ignoring comments" << endl;
#endif
      }
      else if (cmd == "TRAP")
      {
	StrList discard;

	TrapList::cmd(discard, line);

#ifdef CONFIG_DEBUG
        cout << "Added trap: " << line << endl;
#endif
      }
      else if (cmd == "SET")
      {
	std::string varName = FirstWord(line);

	vars.set(varName, line);

#ifdef CONFIG_DEBUG
        cout << "Set: " << varName << " = " << line << endl;
#endif
      }
      else
      {
#ifdef CONFIG_DEBUG
        cout << "Unknown setting type: " << cmd << endl;
#endif
	file.close();
	return false;
      }
    }

    // Done!
    file.close();

    return true;
  }
  else
  {
     return false;
  }
}

