// $Id: config.h,v 1.23 2003/05/17 18:52:26 toast Exp $

#ifndef _CONFIG_H_
#define _CONFIG_H_

#include <string>
#include <list>

#include "strtype"
#include "oomon.h"
#include "botsock.h"

// User flags
enum UserFlags
{
  UF_NONE	= 0x000,
  UF_AUTHED	= 0x001,
  UF_CHANOP	= 0x002,
  UF_OPER	= 0x004,
  UF_WALLOPS	= 0x008,
  UF_CONN	= 0x010,
  UF_NICK	= 0x020,
  UF_KLINE	= 0x040,
  UF_GLINE	= 0x080,
  UF_DLINE	= 0x100,
  UF_REMOTE	= 0x200,
  UF_MASTER	= 0x400,
  UF_ALL	= 0x7FF
};

// Link flags
enum LinkFlags
{
  LF_NONE	= 0x0000
};

// Connect flags
enum ConnectFlags
{
  CF_NONE	= 0x0000
};

class OperDef
{
public:
  std::string Handle, Mask, Passwd;
  int Flags;
  OperDef();
  OperDef(const OperDef & def);
};

class LinkDef
{
public:
  std::string Handle, Mask, Passwd;
  int Flags;
  LinkDef();
  LinkDef(const LinkDef & def);
};

class ConnDef
{
public:
  std::string Handle, Hostname, Passwd;
  int Flags;
  BotSock::Port port;
  ConnDef();
  ConnDef(const ConnDef & def);
};

class YLine
{
public:
  std::string YClass;
  std::string Description;
  YLine();
  YLine(const YLine & def);
};

struct server {
  std::string Hostname, Password, Channels;
  BotSock::Port port;
};

typedef std::list<OperDef> OperList;
typedef std::list<LinkDef> LinkList;
typedef std::list<ConnDef> ConnList;
typedef std::list<YLine> YLineList;

class Config
{
private:
  static OperList Opers;
  static BotSock::Port port;
  static StrList Exceptions;
  static StrList Spoofers;
  static LinkList Links;
  static ConnList Connections;
  static YLineList YLines;
  static std::string Nick, UserName, IRCName, OperNick, OperPass;
  static std::string OurHostName;
  static std::string LogFile;
  static std::string MOTD;
  static struct server Server;
  static std::string proxyVhost;
  static std::string helpFilename;
  static std::string settingsFile;
  static std::string userDBFile;

  static void AddOper(const std::string & Handle, const std::string & Mask,
    const std::string & Passwd, const std::string & Flags);
  static void AddException(const std::string & Mask);
  static void AddSpoofer(const std::string & Mask);
  static void AddLink(const std::string & Handle, const std::string & Mask,
    const std::string & Passwd, const std::string & Flags);
  static void AddConn(const std::string & Handle, const std::string & Hostname,
    const std::string & Passwd, const std::string & Flags,
    const BotSock::Port port);
  static void AddYLine(const std::string & YClass,
    const std::string & Description);
public:
  static void Clear();
  static void Load(const std::string filename = DEFAULT_CFGFILE);
  static bool Auth(const std::string & Handle, const std::string & UserName,
    const std::string & Password, int & Flags, std::string & NewHandle);
  static std::string GetHostName();
  static std::string GetUserName();
  static std::string GetNick() { return Nick; }
  static std::string GetIRCName() { return IRCName; }
  static std::string GetServerHostName() { return Server.Hostname; }
  static BotSock::Port GetServerPort() { return Server.port; }
  static std::string GetServerPassword() { return Server.Password; }
  static std::string GetOperNick() { return OperNick; }
  static std::string GetOperPass() { return OperPass; }
  static std::string GetChannels() { return Server.Channels; }
  static std::string GetLogFile() { return LogFile; }
  static std::string GetMOTD() { return MOTD; }
  static BotSock::Port GetPort() { return port; }
  static bool IsOKHost(const std::string &);
  static bool IsOKHost(const std::string &, const std::string &);
  static bool IsOper(const std::string &);
  static bool IsOper(const std::string &, const std::string &);
  static bool IsLinkable(const std::string &);
  static bool GetConn(std::string &, std::string &, BotSock::Port &,
    std::string &, int &);
  static bool AuthBot(const std::string &, const std::string &,
    const std::string &);
  static std::string GetYLineDescription(const std::string &);
  static std::string getProxyVhost();
  static bool IsSpoofer(const std::string &);
  static std::string getHelpFilename() { return helpFilename; }
  static std::string getUserDBFile() { return userDBFile; }
  static bool saveSettings();
  static bool loadSettings();
};

#endif
