// $Id: dcc.cc,v 1.87 2003/05/25 17:17:50 toast Exp $

// Std C++ Headers
#include <iostream>
#include <fstream>
#include <string>
#include <list>

// Std C Headers

// OOMon Headers
#include "strtype"
#include "oomon.h"
#include "dcc.h"
#include "dcclist.h"
#include "botexcept.h"
#include "config.h"
#include "log.h"
#include "irc.h"
#include "util.h"
#include "main.h"
#include "engine.h"
#include "help.h"
#include "trap.h"
#include "userdb.h"
#include "proxy.h"
#include "seedrand.h"
#include "vars.h"
#include "userhash.h"
#include "remote.h"


#ifdef DEBUG
# define DCC_DEBUG
#endif


DCC::DCC(const std::string & nick, const std::string & userhost)
  : BotSock()
{
  this->setBuffering(true);

  this->echoMyChatter = false;
  this->UserHost = userhost;
  this->Nick = nick;
  this->Flags = UF_NONE;

  this->watches = WatchSet::defaults();
}


DCC::DCC(DCC *listener, const std::string & nick, const std::string & userhost)
  : BotSock(listener)
{
  this->setBuffering(true);

  this->echoMyChatter = listener->echoMyChatter;
  this->UserHost = listener->UserHost;
  this->Nick = listener->Nick;
  this->Flags = listener->Flags;
  this->watches = listener->watches;
}


bool
DCC::connect(const BotSock::Address address, const BotSock::Port port,
  const std::string nick, const std::string userhost)
{
  this->Nick = nick;
  this->UserHost = userhost;

  return this->BotSock::connect(address, port);
}


bool
DCC::listen(const std::string nick, const std::string userhost,
  const BotSock::Port port = 0, const int backlog = 1)
{
  this->Nick = nick;
  this->UserHost = userhost;

  this->setTimeout(60);
  return this->BotSock::listen(port, backlog);
}


// onConnect()
//
// What to do when we connect via DCC with someone
//
// Returns true if no errors ocurred
//
bool
DCC::onConnect()
{
#ifdef DCC_DEBUG
  cout << "DCC::onConnect()" << endl;
#endif
  this->send("OOMon-" + std::string(OOMON_VERSION) + " by Timothy L. Jensen");
  clients.sendAll(this->getUserhost() + " has connected", UF_AUTHED,
    WatchSet(), this);
  this->motd();
  this->setTimeout(DCC_IDLE_MAX);

  return true;
}


// onRead()
//
// Process incoming data from a DCC connection
//
// Returns true if no errors ocurred
//
bool
DCC::onRead(std::string text)
{
#ifdef DCC_DEBUG
  cout << "DCC >> " << text << endl;
#endif

  return this->parse(text);
}


// parse(text)
//
// This is where we actually parse the data received. Fun stuff.
//
// Returns true if the DCC connection should be left open.
//
bool
DCC::parse(std::string text)
{
  std::string Command, To;
  int AtSign;

  if (text == "")
    return true;

  if (text[0] == '.')				// A command?
  {
    // Command
    text.erase((std::string::size_type) 0, 1);
    Command = FirstWord(text);
    AtSign = Command.find('@');  		// Is this a remote command?
    if (AtSign != std::string::npos)
    {
      if ((Flags & UF_REMOTE) == 0)  		// Yes! Do I have permission?
      {
        this->send("*** You can't use remote commands!");
	return true;
      }
      To = Command.substr(AtSign + 1);
      Command = Command.substr(0, AtSign);
      #ifdef DCC_DEBUG
	cout << "Remote command for " << To << ": " << Command << endl;
      #endif
    }
    else
    {
      To = "";
#ifdef DCC_DEBUG
      cout << "Local command: " << Command << endl;
#endif
    }
    if (To != "")
    {
      if (Same(To, Config::GetNick())) 	// No need to route to myself!
      {
        To = "";
      }
      else
      {
	if (!Remote::FindBot(To))
	{
	  this->send("*** Bot not linked: " + To);
	  return true;
        }
      }
    }

    // This is a bit messy...
    switch(DCC::getCommand(UpCase(Command))) {
    case DCC_AUTH:
      if (To == "")
      {
	std::string Parm1 = FirstWord(text);
	if (Config::Auth(Parm1, UserHost, text, Flags, Handle))
	{
	  this->setTimeout(0);
	  this->send("*** Authorization granted!");
	  ::SendAll(Handle + " (" + UserHost + ") has been authorized",
	    UF_AUTHED, WatchSet(), this);
	  Log::Write(Handle + " (" + UserHost + ") has been authorized");

          this->loadConfig();
	}
	else
	{
	  this->send("*** Authorization denied!");
	  ::SendAll(UserHost + " has failed authorization!", UF_AUTHED,
	    WatchSet(), this);
	  Log::Write(UserHost + " has failed authorization!");
	}
      }
      break;
    case DCC_QUIT:
      if (To == "")
        return false;
      break;
    case DCC_CHAT:
      if (To == "")
	this->chat(text);
      break;
    case DCC_WHO:
      if (this->isAuthed())
      {
        if (To == "")
        {
          StrList Output;
	  clients.who(Output);
	  this->send(Output);
        }
        else
        {
          Remote::Who(Handle, To);
        }
      }
      else
      {
	this->notAuthed();
      }
      break;
    case DCC_LINKS:
      if (this->isAuthed())
      {
        if (To == "")
        {
          StrList Output;
	  Remote::GetLinks(Output);
	  this->send(Output);
        }
        else
        {
	  Remote::DoLinks(Handle, To);
        }
      }
      else
      {
	this->notAuthed();
      }
      break;
    case DCC_HELP:
      if (To == "")
        this->help(text);
      break;
    case DCC_WATCH:
      this->watch(To, text);
      break;
    case DCC_ECHO:
      {
	std::string Parm = UpCase(FirstWord(text));
	this->doEcho(Parm);
      }
      break;
    case DCC_OP:
      if (Flags & UF_CHANOP) {
	std::string Parm1 = FirstWord(text);
	std::string Parm2 = FirstWord(text);
	if (Parm1 != "")
	  if (Parm2 != "")
	    server.op(Parm1, Parm2);
	  else
	    server.op(Parm1, Nick);
	else
	  this->send("*** Please include a channel");
      } else
	this->noAccess();
      break;
    case DCC_JOIN:
      if (Flags & UF_CHANOP) {
	std::string Parm1 = FirstWord(text);
	std::string Parm2 = FirstWord(text);
	if (Parm1 != "")
	  server.join(Parm1, Parm2);
	else
	  this->send("*** Please include a channel");
      } else
	this->noAccess();
      break;
    case DCC_PART:
      if (Flags & UF_CHANOP) {
	std::string Parm1 = FirstWord(text);
	if (Parm1 != "")
	  server.part(Parm1);
	else
	  this->send("*** Please include a channel");
      } else
	this->noAccess();
      break;
    case DCC_KLINE:
      if (Flags & UF_KLINE) {
	unsigned int Minutes = 0;
	std::string Parm1 = FirstWord(text);
	if ((Minutes = atoi(Parm1.c_str())) != 0)
	  Parm1 = FirstWord(text);
	if (Parm1 != "")
	  if (text != "")
	    if (To == "")
	      this->kline(Minutes, Parm1, text);
	    else
	      Remote::Kline(Handle, To, Minutes, Parm1, text);
	  else
	    this->send("*** Please include a kline reason");
	else
	  this->send("*** Please include a nick or user@host mask to kline");
      } else
	this->noAccess();
      break;
    case DCC_UNKLINE:
      if (Flags & UF_KLINE) {
	std::string Parm1 = FirstWord(text);
	if (Parm1 != "")
	  if (To == "")
	    this->unkline(Parm1);
	  else
	    Remote::Unkline(Handle, To, Parm1);
	else
	  this->send("*** Please include a user@host mask to unkline");
      } else
	this->noAccess();
      break;
    case DCC_KCLONE:
      if (Flags & UF_KLINE) {
	unsigned int Minutes = 0;
	std::string Parm1 = FirstWord(text);
	if ((Minutes = atoi(Parm1.c_str())) != 0)
	  Parm1 = FirstWord(text);
	if (Parm1 != "")
	  if (To == "")
	    this->kClone(Minutes, Parm1);
	  else
	    Remote::KClone(Handle, To, Minutes, Parm1);
	else
	  this->send("*** Please include a nick or user@host mask to kline");
      } else
	this->noAccess();
      break;
    case DCC_KFLOOD:
      if (Flags & UF_KLINE) {
	unsigned int Minutes = 0;
	std::string Parm1 = FirstWord(text);
	if ((Minutes = atoi(Parm1.c_str())) != 0)
	  Parm1 = FirstWord(text);
	if (Parm1 != "")
	  if (To == "")
	    this->kFlood(Minutes, Parm1);
	  else
	    Remote::KFlood(Handle, To, Minutes, Parm1);
	else
	  this->send("*** Please include a nick or user@host mask to kline");
      } else
	this->noAccess();
      break;
    case DCC_KBOT:
      if (Flags & UF_KLINE) {
	unsigned int Minutes = 0;
	std::string Parm1 = FirstWord(text);
	if ((Minutes = atoi(Parm1.c_str())) != 0)
	  Parm1 = FirstWord(text);
	if (Parm1 != "")
	  if (To == "")
	    this->kBot(Minutes, Parm1);
	  else
	    Remote::KBot(Handle, To, Minutes, Parm1);
	else
	  this->send("*** Please include a nick or user@host mask to kline");
      } else
	this->noAccess();
      break;
    case DCC_KLINK:
      if (Flags & UF_KLINE) {
	unsigned int Minutes = 0;
	std::string Parm1 = FirstWord(text);
	if ((Minutes = atoi(Parm1.c_str())) != 0)
	  Parm1 = FirstWord(text);
	if (Parm1 != "")
	  if (To == "")
	    this->kLink(Minutes, Parm1);
	  else
	    Remote::KLink(Handle, To, Minutes, Parm1);
	else
	  this->send("*** Please include a nick or user@host mask to kline");
      } else
	this->noAccess();
      break;
    case DCC_KSPAM:
      if (Flags & UF_KLINE) {
	unsigned int Minutes = 0;
	std::string Parm1 = FirstWord(text);
	if ((Minutes = atoi(Parm1.c_str())) != 0)
	  Parm1 = FirstWord(text);
	if (Parm1 != "")
	  if (To == "")
	    this->kSpam(Minutes, Parm1);
	  else
	    Remote::KSpam(Handle, To, Minutes, Parm1);
	else
	  this->send("*** Please include a nick or user@host mask to kline");
      } else
	this->noAccess();
      break;
    case DCC_KTRACE:
      if (Flags & UF_KLINE) {
	unsigned int Minutes = 0;
	std::string Parm1 = FirstWord(text);
	if ((Minutes = atoi(Parm1.c_str())) != 0)
	  Parm1 = FirstWord(text);
	if (Parm1 != "")
	  if (To == "")
	    this->kTrace(Minutes, Parm1);
	  else
	    Remote::KTrace(Handle, To, Minutes, Parm1);
	else
	  this->send("*** Please include a nick or user@host mask to kline");
      } else
	this->noAccess();
      break;
    case DCC_KMOTD:
      if (Flags & UF_KLINE) {
	unsigned int Minutes = 0;
	std::string Parm1 = FirstWord(text);
	if ((Minutes = atoi(Parm1.c_str())) != 0)
	  Parm1 = FirstWord(text);
	if (Parm1 != "")
	  if (To == "")
	    this->kMotd(Minutes, Parm1);
	  else
	    Remote::KMotd(Handle, To, Minutes, Parm1);
	else
	  this->send("*** Please include a nick or user@host mask to kline");
      } else
	this->noAccess();
      break;
    case DCC_KINFO:
      if (Flags & UF_KLINE) {
	unsigned int Minutes = 0;
	std::string Parm1 = FirstWord(text);
	if ((Minutes = atoi(Parm1.c_str())) != 0)
	  Parm1 = FirstWord(text);
	if (Parm1 != "")
	  if (To == "")
	    this->kInfo(Minutes, Parm1);
	  else
	    Remote::KInfo(Handle, To, Minutes, Parm1);
	else
	  this->send("*** Please include a nick or user@host mask to kline");
      } else
	this->noAccess();
      break;
    case DCC_KPROXY:
      if (Flags & UF_KLINE) {
	unsigned int Minutes = 0;
	std::string Parm1 = FirstWord(text);
	if ((Minutes = atoi(Parm1.c_str())) != 0)
	  Parm1 = FirstWord(text);
	if (Parm1 != "")
	  if (To == "")
	    this->kProxy(Minutes, Parm1);
	  else
	    Remote::KProxy(Handle, To, Minutes, Parm1);
	else
	  this->send("*** Please include a nick or user@host mask to kline");
      } else
	this->noAccess();
      break;
    case DCC_KPERM:
      if (Flags & UF_KLINE) {
	std::string Parm1 = FirstWord(text);
	if (Parm1 != "")
	  if (text != "")
	    if (To == "")
	      this->kPerm(Parm1, text);
	    else
	      Remote::KPerm(Handle, To, Parm1, text);
	  else
	    this->send("*** Please include a kline reason");
	else
	  this->send("*** Please include a nick or user@host mask to kline");
      } else
	this->noAccess();
      break;
    case DCC_DLINE:
      if (Flags & UF_DLINE) {
	std::string Parm1 = FirstWord(text);
	if (Parm1 != "")
	  if (text != "")
	    if (To == "")
	      this->dline(Parm1, text);
	    else
	      Remote::Dline(Handle, To, Parm1, text);
	  else
	    this->send("*** Please include a reason for the dline");
	else
	  this->send("*** Please include an IP mask to dline");
      } else
	this->noAccess();
      break;
    case DCC_UNDLINE:
      if (Flags & UF_DLINE) {
	std::string Parm1 = FirstWord(text);
	if (Parm1 != "")
	  if (To == "")
	    this->undline(Parm1);
	  else
	    Remote::Undline(Handle, To, Parm1);
	else
	  this->send("*** Please include a user@host mask to undline");
      } else
	this->noAccess();
      break;
    case DCC_GLINE:
      if (Flags & UF_GLINE) {
	if (To == "") {
	  unsigned int Minutes = 0;
	  std::string Parm1 = FirstWord(text);
	  if ((Minutes = atoi(Parm1.c_str())) != 0)
	    Parm1 = FirstWord(text);
	  if (Parm1 != "")
	    if (text != "")
	      this->gline(Minutes, Parm1, text);
	    else
	      this->send("*** Please include a gline reason");
	  else
	    this->send("*** Please include a nick or user@host mask to gline");
	} else
	  this->notRemote();
      } else
	this->noAccess();
      break;
    case DCC_UNGLINE:
      if (Flags & UF_GLINE) {
	if (To == "") {
	  std::string Parm1 = FirstWord(text);
	  if (Parm1 != "")
	    this->ungline(Parm1);
	  else
	    this->send("*** Please include a user@host mask to gline");
	} else
	  this->notRemote();
      } else
	this->noAccess();
      break;
    case DCC_GCLONE:
      if (Flags & UF_GLINE) {
	if (To == "") {
	  unsigned int Minutes = 0;
	  std::string Parm1 = FirstWord(text);
	  if ((Minutes = atoi(Parm1.c_str())) != 0)
	    Parm1 = FirstWord(text);
	  if (Parm1 != "")
	    this->gline(Minutes, Parm1, REASON_CLONES);
	  else
	    this->send("*** Please include a nick or user@host mask to gline");
	} else
	  this->notRemote();
      } else
	this->noAccess();
      break;
    case DCC_GFLOOD:
      if (Flags & UF_GLINE) {
	if (To == "") {
	  unsigned int Minutes = 0;
	  std::string Parm1 = FirstWord(text);
	  if ((Minutes = atoi(Parm1.c_str())) != 0)
	    Parm1 = FirstWord(text);
	  if (Parm1 != "")
	    this->gline(Minutes, Parm1, REASON_FLOODING);
	  else
	    this->send("*** Please include a nick or user@host mask to gline");
	} else
	  this->notRemote();
      } else
	this->noAccess();
      break;
    case DCC_GBOT:
      if (Flags & UF_GLINE) {
	if (To == "") {
	  unsigned int Minutes = 0;
	  std::string Parm1 = FirstWord(text);
	  if ((Minutes = atoi(Parm1.c_str())) != 0)
	    Parm1 = FirstWord(text);
	  if (Parm1 != "")
	    this->gline(Minutes, Parm1, REASON_BOTS);
	  else
	    this->send("*** Please include a nick or user@host mask to gline");
	} else
	  this->notRemote();
      } else
	this->noAccess();
      break;
    case DCC_GLINK:
      if (Flags & UF_GLINE) {
	if (To == "") {
	  unsigned int Minutes = 0;
	  std::string Parm1 = FirstWord(text);
	  if ((Minutes = atoi(Parm1.c_str())) != 0)
	    Parm1 = FirstWord(text);
	  if (Parm1 != "")
	    this->gline(Minutes, Parm1, REASON_LINKS);
	  else
	    this->send("*** Please include a nick or user@host mask to gline");
	} else
	  this->notRemote();
      } else
	this->noAccess();
      break;
    case DCC_GSPAM:
      if (Flags & UF_GLINE) {
	if (To == "") {
	  unsigned int Minutes = 0;
	  std::string Parm1 = FirstWord(text);
	  if ((Minutes = atoi(Parm1.c_str())) != 0)
	    Parm1 = FirstWord(text);
	  if (Parm1 != "")
	    this->gline(Minutes, Parm1, REASON_SPAM);
	  else
	    this->send("*** Please include a nick or user@host mask to gline");
	} else
	  this->notRemote();
      } else
	this->noAccess();
      break;
    case DCC_GTRACE:
      if (Flags & UF_GLINE) {
	if (To == "") {
	  unsigned int Minutes = 0;
	  std::string Parm1 = FirstWord(text);
	  if ((Minutes = atoi(Parm1.c_str())) != 0)
	    Parm1 = FirstWord(text);
	  if (Parm1 != "")
	    this->gline(Minutes, Parm1, REASON_TRACE);
	  else
	    this->send("*** Please include a nick or user@host mask to gline");
	} else
	  this->notRemote();
      } else
	this->noAccess();
      break;
    case DCC_GMOTD:
      if (Flags & UF_GLINE) {
	if (To == "") {
	  unsigned int Minutes = 0;
	  std::string Parm1 = FirstWord(text);
	  if ((Minutes = atoi(Parm1.c_str())) != 0)
	    Parm1 = FirstWord(text);
	  if (Parm1 != "")
	    this->gline(Minutes, Parm1, REASON_MOTD);
	  else
	    this->send("*** Please include a nick or user@host mask to gline");
	} else
	  this->notRemote();
      } else
	this->noAccess();
      break;
    case DCC_GINFO:
      if (Flags & UF_GLINE) {
	if (To == "") {
	  unsigned int Minutes = 0;
	  std::string Parm1 = FirstWord(text);
	  if ((Minutes = atoi(Parm1.c_str())) != 0)
	    Parm1 = FirstWord(text);
	  if (Parm1 != "")
	    this->gline(Minutes, Parm1, REASON_INFO);
	  else
	    this->send("*** Please include a nick or user@host mask to gline");
	} else
	  this->notRemote();
      } else
	this->noAccess();
      break;
    case DCC_GPROXY:
      if (Flags & UF_GLINE) {
	if (To == "") {
	  unsigned int Minutes = 0;
	  std::string Parm1 = FirstWord(text);
	  if ((Minutes = atoi(Parm1.c_str())) != 0)
	    Parm1 = FirstWord(text);
	  if (Parm1 != "")
	    this->gline(Minutes, Parm1, REASON_PROXY);
	  else
	    this->send("*** Please include a nick or user@host mask to gline");
	} else
	  this->notRemote();
      } else
	this->noAccess();
      break;
    case DCC_GPERM:
      if (Flags & UF_GLINE) {
	if (To == "") {
	  std::string Parm1 = FirstWord(text);
	  if (Parm1 != "")
	    if (text != "")
	      this->gline(0, Parm1, "PERM " + text);
	    else
	      this->send("*** Please include a gline reason");
	  else
	    this->send("*** Please include a nick or user@host mask to gline");
	} else
	  this->notRemote();
      } else
	this->noAccess();
      break;
    case DCC_KILL:
      if (Flags & UF_OPER) {
	if (To == "") {
	  std::string Parm1 = FirstWord(text);
	  if (Parm1 != "") {
	    if (text != "") {
	      ::SendAll("KILL " + Parm1 + " ((" + this->getHandle() + ") " +
		text + ")", UF_OPER, WATCH_KILLS, this);
	      server.kill(this->getHandle(), Parm1, text);
	    } else
	      this->send("*** Please include a kill reason.");
	  } else
	    this->send("*** Please include a nick to kill.");
	}
      } else
	this->noAccess();
      break;
    case DCC_KILLLIST:
      this->killList(To, text);
      break;
    case DCC_MULTI:
      if (Flags & UF_OPER) {
	if (To == "") {
	  std::string Parm1 = FirstWord(text);
	  int min = atoi(Parm1.c_str());
	  StrList Output;
	  users.reportMulti(Output, min);
	  this->send(Output);
	}
      } else
	this->noAccess();
      break;
    case DCC_HMULTI:
      if (Flags & UF_OPER) {
	if (To == "") {
	  std::string Parm1 = FirstWord(text);
	  int min = atoi(Parm1.c_str());
	  StrList Output;
	  users.reportHMulti(Output, min);
	  this->send(Output);
	}
      } else
	this->noAccess();
      break;
    case DCC_UMULTI:
      if (Flags & UF_OPER) {
	if (To == "") {
	  std::string Parm1 = FirstWord(text);
	  int min = atoi(Parm1.c_str());
	  StrList Output;
	  users.reportUMulti(Output, min);
	  this->send(Output);
	}
      } else
	this->noAccess();
      break;
    case DCC_VMULTI:
      if (Flags & UF_OPER) {
	if (To == "") {
	  std::string Parm1 = FirstWord(text);
	  int min = atoi(Parm1.c_str());
	  StrList Output;
	  users.reportVMulti(Output, min);
	  this->send(Output);
	}
      } else
	this->noAccess();
      break;
    case DCC_NFIND:
      this->nFind(To, text);
      break;
    case DCC_TRACE:
      if (Flags & UF_OPER) {
	if (To == "") {
	  server.retrace(this->getHandle());
	}
      } else
	this->noAccess();
      break;
    case DCC_CLONES:
      if (Flags & UF_OPER) {
	if (To == "") {
	  StrList Output;
	  users.reportClones(Output);
	  this->send(Output);
	}
      } else
	this->noAccess();
      break;
    case DCC_LIST:
      this->list(To, text);
      break;
    case DCC_FINDK:
      this->findK(To, text);
      break;
    case DCC_FINDD:
      this->findD(To, text);
      break;
    case DCC_RELOAD:
      if (Flags & UF_OPER) {
	if (To == "") {
	  std::string Parm1 = UpCase(FirstWord(text));
	  if (Parm1 == "USERS") {
	    server.retrace(this->getHandle());
	  } else if (Parm1 == "KLINES") {
	    server.reloadKlines(this->getHandle());
	  } else if (Parm1 == "DLINES") {
	    server.reloadDlines(this->getHandle());
	  } else if (Parm1 == "CONFIG") {
	    ReloadConfig(this->getHandle());
	  } else if (Parm1 == "HELP") {
	    Help::flush();
	  } else
	    this->send("*** Valid options are USERS, KLINES, CONFIG, and HELP");
	}
      } else
	this->noAccess();
      break;
    case DCC_DOMAINS:
      if (Flags & UF_OPER) {
	if (To == "") {
          std::string Parm1 = FirstWord(text);
	  int min = 5;
	  if (Parm1 != "")
	    min = atoi(Parm1.c_str());
	  if (min >= 1)
	  {
	    StrList Output;
	    users.reportDomains(Output, min);
	    this->send(Output);
	  } else
	    this->send("*** Invalid minimum user size!");
	}
      } else
	this->noAccess();
      break;
    case DCC_CLASS:
      if (Flags & UF_OPER) {
	if (To == "") {
          std::string className = FirstWord(text);
	  StrList Output;
	  users.reportClasses(Output, className);
	  this->send(Output);
	}
      } else
	this->noAccess();
      break;
    case DCC_MOTD:
      if (this->isAuthed())
      {
	if (To == "")
	{
	  this->motd();
	}
      }
      else
      {
	this->notAuthed();
      }
      break;
    case DCC_CONN:
      if (Flags & UF_MASTER) {
	std::string Parm1 = FirstWord(text);
	if (Parm1 != "")
	  Remote::Conn(Handle, To, Parm1);
	else
	  this->send("*** Please specify the bot to connect");
      } else
	this->noAccess();
      break;
    case DCC_DISCONN:
      if (Flags & UF_MASTER) {
	std::string Parm1 = FirstWord(text);
	if (Parm1 != "")
	  Remote::Disconn(Handle, To, Parm1);
	else
	  this->send("*** Please specify the bot to disconnect");
      } else
	this->noAccess();
      break;
    case DCC_DIE:
      if (Flags & UF_MASTER) {
	if (To == "") {
	  ::SendAll(this->getHandle() + " says I have to die now...", UF_NONE,
	    WatchSet(), this);
	  this->send("As you wish...");
	  Log::Write("DIE requested by " + this->getHandle());
	  gracefuldie(SIGTERM); 
	} else
	  Remote::Die(Handle, To);
      } else
	this->noAccess();
      break;
    case DCC_RAW:
      if (Flags & UF_MASTER) {
	if (To == "") {
	  Log::Write("RAW by " + this->getHandle() + ": " + text);
	  server.write(text + "\n");
	}
      } else
	this->noAccess();
      break;
    case DCC_TRAP:
      this->trap(To, text);
      break;
    case DCC_SEEDRAND:
      this->seedrand(To, text);
      break;
    case DCC_STATUS:
      this->status(To, text);
      break;
    case DCC_SAVE:
      this->save(To, text);
      break;
    case DCC_LOAD:
      this->load(To, text);
      break;
    case DCC_SET:
      this->set(To, text);
      break;
    case DCC_SPAMSUB:
      this->spamsub(To, text);
      break;
    case DCC_SPAMUNSUB:
      this->spamunsub(To, text);
      break;
#ifdef DEBUG
    case DCC_TEST:
      if (Flags & UF_MASTER)
      {
	std::string score = IntToStr(seedrandScore(FirstWord(text)));
	this->send("Score = " + score);
      }
      else
      {
	this->noAccess();
      }
      break;
#endif
    case DCC_UNKNOWN:
      this->send("*** Invalid command: " + Command);
      break;
    }
  }
  else					// Just chatter :P
  {
    // Chat
    if (text == "Unknown command" &&
      vars[VAR_IGNORE_UNKNOWN_COMMAND]->getBool())
    {
      // Just ignore it.
      return true;
    }
    this->chat(text);
  }
  return true;
}


// chat(text)
//
// Handles any text which should be interpretted as chatter.
//
void
DCC::chat(const std::string & text)
{
  if (this->isAuthed() || vars[VAR_UNAUTHED_MAY_CHAT]->getBool())
  {
    clients.chat(this->getHandle(), "*", text, echoMyChatter ? NULL : this);
    Remote::Chat(this->getHandle(), "*", text);
  }
  else
  {
    this->send("*** You must authenticate yourself with \".auth\" before you can chat!");
  }
}


// motd()
//
// Writes the Message Of The Day
//
void
DCC::motd()
{
  StrList Output;
  ::motd(Output);
  this->send(Output);
}


// noAccess()
//
// Tells the user that he doesn't have access to use the command he passed
//
void
DCC::noAccess()
{
  this->send("*** You don't have access for that command!");
}


// notRemote()
//
// Tells the user that the command he used isn't usable remotely.
//
void
DCC::notRemote()
{
  this->send("*** That command can't be used remotely!");
}


// notAuthed()
//
// Tells the user that the command he must be authenticated before he can
// use the command.
//
void
DCC::notAuthed()
{
  this->send("*** You must authenticate yourself with \".auth\" before you can use that command!");
}


std::string
DCC::getHandle() const
{
  return (this->Handle + "@" + Config::GetNick());
}


void
DCC::list(const std::string & to, std::string text)
{
  if (this->Flags & UF_OPER)
  {
    if (to == "")
    {
      std::string parm = FirstWord(text);
      std::string className, mask;

      if (Same(parm, "-CLASS"))
      {
        className = FirstWord(text);
        mask = FirstWord(text);
      }
      else
      {
        mask = parm;
      }

      if (mask != "")
      {
        StrList Output;
        users.listUsers(Output, mask, className);
        this->send(Output);
      }
      else
      {
        this->send("*** Syntax: .list [-class <name>] <mask>");
      }
    }
    else
    {
      this->notRemote();
    }
  }
  else
  {
    this->noAccess();
  }
}


void
DCC::killList(const std::string & to, std::string text)
{
  if (this->Flags & UF_OPER)
  {
    if (to == "")
    {
      std::string parm = FirstWord(text);
      std::string className, mask;

      if (Same(parm, "-CLASS"))
      {
        className = FirstWord(text);
        mask = FirstWord(text);
      }
      else
      {
        mask = parm;
      }

      if (mask != "")
      {
        StrList Output;
        if (text == "")
        {
          ::SendAll("KILLLIST " + mask + " [" + this->getHandle() + "]",
            UF_OPER, WATCH_KILLS, this);
          users.listUsers(Output, mask, className, true, this->getHandle(),
	    vars[VAR_KILLLIST_REASON]->getString());
        }
        else
        {
          ::SendAll("KILLLIST " + mask + " (" + text + ") [" +
            this->getHandle() + "]", UF_OPER, WATCH_KILLS, this);
	  users.listUsers(Output, mask, className, true, this->getHandle(),
            text);
	}
      }
      else
      {
        this->send("*** Syntax: .killlist [-class <name>] <mask>");
      }
    }
    else
    {
      this->notRemote();
    }
  }
  else
  {
    this->noAccess();
  }
}


void
DCC::nFind(const std::string & to, std::string text)
{
  if (this->Flags & UF_OPER)
  {
    if (to == "")
    {
      std::string parm = FirstWord(text);
      std::string className, mask;

      if (Same(parm, "-CLASS"))
      {
        className = FirstWord(text);
        mask = FirstWord(text);
      }
      else
      {
        mask = parm;
      }

      if (mask != "")
      {
        StrList Output;
        users.listNicks(Output, parm, className);
        this->send(Output);
      }
      else
      {
        this->send("*** Syntax: .nfind [-class <name>] <mask>");
      }
    }
    else
    {
      this->notRemote();
    }
  }
  else
  {
    this->noAccess();
  }
}


void
DCC::findK(const std::string & to, std::string text)
{
  if (this->Flags & UF_OPER)
  {
    if (to == "")
    {
      bool count = false;
      bool remove = false;
      std::string parm = FirstWord(text);
      std::string mask;
      if (Same(parm, "-COUNT"))
      {
        count = true;
        mask = FirstWord(text);
      }
      else if (Same(parm, "-REMOVE"))
      {
        if (Flags & UF_KLINE)
        {
          remove = true;
          mask = FirstWord(text);
        }
        else
        {
          this->noAccess();
          return;
        }
      }
      else
      {
        mask = parm;
      }

      if (mask != "")
      {
        if (remove)
        {
          int removed = server.findAndRemoveK(this->getHandle(), mask);
          this->send("*** " + IntToStr(removed) + " K-lines removed.");
        }
        else
        {
          StrList Output;
          server.findK(Output, mask, count);
          this->send(Output);
        }
      }
      else
      {
        this->send("*** Syntax: .findk [-COUNT|-REMOVE] <mask>");
      }
    }
    else
    {
      this->notRemote();
    }
  }
  else
  {
    this->noAccess();
  }
}


void
DCC::findD(const std::string & to, std::string text)
{
  if (this->Flags & UF_OPER)
  {
    if (to == "")
    {
      bool count = false;
      bool remove = false;
      std::string parm = FirstWord(text);
      std::string mask;
      if (Same(parm, "-COUNT"))
      {
        count = true;
        mask = FirstWord(text);
      }
      else if (Same(parm, "-REMOVE"))
      {
        if (Flags & UF_DLINE)
        {
          remove = true;
          mask = FirstWord(text);
        }
        else
        {
          this->noAccess();
          return;
        }
      }
      else
      {
        mask = parm;
      }
      if (mask != "")
      {
        if (remove)
        {
          int removed = server.findAndRemoveD(this->getHandle(), mask);
          this->send("*** " + IntToStr(removed) + " D-lines removed.");
        }
        else
        {
          StrList Output;
          server.findD(Output, mask, count);
          this->send(Output);
        }
      }
      else
      {
        this->send("*** Syntax: .findd [-COUNT|-REMOVE] <mask>");
      }
    }
    else
    {
      this->notRemote();
    }
  }
  else
  {
    this->noAccess();
  }
}


void
DCC::seedrand(const std::string & to, std::string text)
{
  if (this->Flags & UF_OPER)
  {
    if (to == "")
    {
      StrList output;
      std::string mask = FirstWord(text);
      if (mask != "")
      {
	int threshold = atoi(FirstWord(text).c_str());
	if (threshold > 0)
	{
	  users.reportSeedrand(output, mask, threshold);
	}
	else
	{
	  users.reportSeedrand(output, mask,
	    vars[VAR_SEEDRAND_COMMAND_MIN]->getInt());
	}
      }
      else
      {
	users.reportSeedrand(output, "*",
	  vars[VAR_SEEDRAND_COMMAND_MIN]->getInt());
      }
      this->send(output);
    }
    else
    {
      this->notRemote();
    }
  }
  else
  {
    this->noAccess();
  }
}


void
DCC::status(const std::string & to, std::string text)
{
  if (this->Flags & UF_MASTER)
  {
    if (to == "")
    {
      StrList output;
      users.status(output);
      ::status(output);
      server.status(output);
      clients.status(output);
      Proxy::status(output);
      this->send(output);
    }
    else
    {
      this->notRemote();
    }
  }
  else
  {
    this->noAccess();
  }
}


void
DCC::save(const std::string & to, std::string text)
{
  if (this->Flags & UF_MASTER)
  {
    if (to == "")
    {
      if (Config::saveSettings())
      {
        std::string notice("*** Saved settings to file.");
        ::SendAll(notice, UF_OPER);
        Log::Write(notice);
      }
      else
      {
	this->send("*** Error saving settings to file!");
      }
    }
    else
    {
      this->notRemote();
    }
  }
  else
  {
    this->noAccess();
  }
}


void
DCC::load(const std::string & to, std::string text)
{
  if (this->Flags & UF_MASTER)
  {
    if (to == "")
    {
      if (Config::loadSettings())
      {
        std::string notice("*** Loaded settings from file.");
        ::SendAll(notice, UF_OPER);
        Log::Write(notice);
      }
      else
      {
	this->send("*** Error loading settings from file!");
      }
    }
    else
    {
      this->notRemote();
    }
  }
  else
  {
    this->noAccess();
  }
}


void
DCC::set(const std::string & to, std::string text)
{
  if (this->Flags & UF_MASTER)
  {
    if (to == "")
    {
      std::string varName = FirstWord(text);
      bool clearVar = false;

      if ((varName.length() > 1) && (varName[0] == '-'))
      {
        clearVar = true;
	varName = varName.substr(1);
      }

      if ((text.length() == 0) && !clearVar)
      {
	StrList output;
	if (vars.get(output, varName) > 0)
	{
	  this->send(output);
	}
	else
	{
	  this->send("No such variable \"" + varName + "\"");
	}
      }
      else
      {
	std::string error = vars.set(varName, text, this->getHandle());
	if (error.length() > 0)
	{
	  this->send(error);
	}
      }
    }
    else
    {
      this->notRemote();
    }
  }
  else
  {
    this->noAccess();
  }
}


void
DCC::spamsub(const std::string & to, std::string text)
{
  if (this->Flags & UF_MASTER)
  {
    if (to == "")
    {
      std::string notice("*** " + this->getHandle() +
	" subscribing to SpamTrap");

      ::SendAll(notice, UF_OPER);
      Log::Write(notice);

      server.subSpamTrap(true);
    }
    else
    {
      this->notRemote();
    }
  }
  else
  {
    this->noAccess();
  }
}


void
DCC::spamunsub(const std::string & to, std::string text)
{
  if (this->Flags & UF_MASTER)
  {
    if (to == "")
    {
      std::string notice("*** " + this->getHandle() +
	" unsubscribing from SpamTrap");

      ::SendAll(notice, UF_OPER);
      Log::Write(notice);

      server.subSpamTrap(false);
    }
    else
    {
      this->notRemote();
    }
  }
  else
  {
    this->noAccess();
  }
}


// kline(Minutes, Target, Reason)
//
// Writes a kline to the server
//
void
DCC::kline(const int Minutes, const std::string & Target,
  const std::string & Reason)
{
  if (Minutes)
    ::SendAll("KLINE " + IntToStr(Minutes) + " " + Target + " (" + Reason +
      " [" + this->getHandle() + "])", UF_OPER, WATCH_KLINES,
      this);
  else
    ::SendAll("KLINE " + Target + " (" + Reason + " [" + this->getHandle() +
      "])", UF_OPER, WATCH_KLINES, this);
  server.kline(this->getHandle(), Minutes, Target, Reason);
}


// kClone(Minutes, Target)
//
// Writes a kline to the server
//
void
DCC::kClone(const int Minutes, const std::string & Target)
{
  if (Minutes)
    ::SendAll("KCLONE " + IntToStr(Minutes) + " " + Target + " [" +
      this->getHandle() + "]", UF_OPER, WATCH_KLINES, this);
  else
    ::SendAll("KCLONE " + Target + " [" + this->getHandle() + "]", UF_OPER,
      WATCH_KLINES, this);
  server.kline(this->getHandle(), Minutes, Target, REASON_CLONES);
}


// kFlood(Minutes, Target)
//
// Writes a kline to the server
//
void
DCC::kFlood(const int Minutes, const std::string & Target)
{
  if (Minutes)
    ::SendAll("KFLOOD " + IntToStr(Minutes) + " " + Target + " [" +
      this->getHandle() + "]", UF_OPER, WATCH_KLINES, this);
  else
    ::SendAll("KFLOOD " + Target + " [" + this->getHandle() + "]", UF_OPER,
      WATCH_KLINES, this);
  server.kline(this->getHandle(), Minutes, Target, REASON_FLOODING);
}


// kBot(Minutes, Target)
//
// Writes a kline to the server
//
void
DCC::kBot(const int Minutes, const std::string & Target)
{
  if (Minutes)
    ::SendAll("KBOT " + IntToStr(Minutes) + " " + Target + " [" +
      this->getHandle() + "]", UF_OPER, WATCH_KLINES, this);
  else
    ::SendAll("KBOT " + Target + " [" + this->getHandle() + "]", UF_OPER,
      WATCH_KLINES, this);
  server.kline(this->getHandle(), Minutes, Target, REASON_BOTS);
}


// kSpam(Minutes, Target)
//
// Writes a kline to the server
//
void
DCC::kSpam(const int Minutes, const std::string & Target)
{
  if (Minutes)
    ::SendAll("KSPAM " + IntToStr(Minutes) + " " + Target + " [" +
      this->getHandle() + "]", UF_OPER, WATCH_KLINES, this);
  else
    ::SendAll("KSPAM " + Target + " [" + this->getHandle() + "]", UF_OPER,
      WATCH_KLINES, this);
  server.kline(this->getHandle(), Minutes, Target, REASON_SPAM);
}


// kLink(Minutes, Target)
//
// Writes a kline to the server
//
void
DCC::kLink(const int Minutes, const std::string & Target)
{
  if (Minutes)
    ::SendAll("KLINK " + IntToStr(Minutes) + " " + Target + " [" +
      this->getHandle() + "]", UF_OPER, WATCH_KLINES, this);
  else
    ::SendAll("KLINK " + Target + " [" + this->getHandle() + "]", UF_OPER,
      WATCH_KLINES, this);
  server.kline(this->getHandle(), Minutes, Target, REASON_LINKS);
}


// kTrace(Minutes, Target)
//
// Writes a kline to the server
//
void
DCC::kTrace(const int Minutes, const std::string & Target)
{
  if (Minutes)
    ::SendAll("KTRACE " + IntToStr(Minutes) + " " + Target + " [" +
      this->getHandle() + "]", UF_OPER, WATCH_KLINES, this);
  else
    ::SendAll("KTRACE " + Target + " [" + this->getHandle() + "]", UF_OPER,
      WATCH_KLINES, this);
  server.kline(this->getHandle(), Minutes, Target, REASON_TRACE);
}


// kMotd(Minutes, Target)
//
// Writes a kline to the server
//
void
DCC::kMotd(const int Minutes, const std::string & Target)
{
  if (Minutes)
    ::SendAll("KMOTD " + IntToStr(Minutes) + " " + Target + " [" +
      this->getHandle() + "]", UF_OPER, WATCH_KLINES, this);
  else
    ::SendAll("KMOTD " + Target + " [" + this->getHandle() + "]", UF_OPER,
      WATCH_KLINES, this);
  server.kline(this->getHandle(), Minutes, Target, REASON_MOTD);
}


// kInfo(Minutes, Target)
//
// Writes a kline to the server
//
void
DCC::kInfo(const int Minutes, const std::string & Target)
{
  if (Minutes)
    ::SendAll("KINFO " + IntToStr(Minutes) + " " + Target + " [" +
      this->getHandle() + "]", UF_OPER, WATCH_KLINES, this);
  else
    ::SendAll("KINFO " + Target + " [" + this->getHandle() + "]", UF_OPER,
      WATCH_KLINES, this);
  server.kline(this->getHandle(), Minutes, Target, REASON_INFO);
}


// kProxy(Minutes, Target)
//
// Writes a kline to the server
//
void
DCC::kProxy(const int Minutes, const std::string & Target)
{
  if (Minutes)
    ::SendAll("KPROXY " + IntToStr(Minutes) + " " + Target + " [" +
      this->getHandle() + "]", UF_OPER, WATCH_KLINES, this);
  else
    ::SendAll("KPROXY " + Target + " [" + this->getHandle() + "]", UF_OPER,
      WATCH_KLINES, this);
  server.kline(this->getHandle(), Minutes, Target, REASON_PROXY);
}


// kPerm(Target, Reason)
//
// Writes a kline to the server
//
void
DCC::kPerm(const std::string & Target, const std::string & Reason)
{
  ::SendAll("KPERM " + Target + " (" + Reason + " [" + this->getHandle() + "])",
    UF_OPER, WATCH_KLINES, this);
  server.kline(this->getHandle(), 0, Target, "PERM " + Reason);
}


// unkline(Target)
//
// Just the opposite of Kline
//
void
DCC::unkline(const std::string & Target)
{
  ::SendAll("UNKLINE " + Target + " [" + this->getHandle() + "]", UF_OPER,
    WATCH_KLINES, this);
  server.unkline(this->getHandle(), Target);
}


// dline(Target)
//
// Writes a dline to the server
//
void
DCC::dline(const std::string & Target, const std::string & Reason)
{
  ::SendAll("DLINE " + Target + " (" + Reason + " [" + this->getHandle() + "])",
    UF_OPER, WATCH_DLINES, this);
  server.dline(this->getHandle(), Target, Reason);
}


// undline(Target)
//
// Just the opposite of dline
//
void
DCC::undline(const std::string & Target)
{
  ::SendAll("UNDLINE " + Target + " [" + this->getHandle() + "]", UF_OPER,
    WATCH_DLINES, this);
  server.undline(this->getHandle(), Target);
}


// gline(Minutes, Target, Reason)
//
// Writes a kline too all linked servers
//
void
DCC::gline(const int Minutes, const std::string & Target,
  const std::string & Reason)
{
  if (Minutes)
    ::SendAll("GLINE " + IntToStr(Minutes) + " " + Target + " (" + Reason +
      " [" + this->getHandle() + "])", UF_OPER, WATCH_KLINES, this);
  else
    ::SendAll("GLINE " + Target + " (" + Reason + " [" + this->getHandle() +
      "])", UF_OPER, WATCH_KLINES, this);
  server.kline(this->getHandle(), Minutes, Target, Reason);
  Remote::Gline(this->getHandle(), Minutes, Target, Reason);
}


// ungline(Target)
//
// Opposite of gline
//
void
DCC::ungline(const std::string & Target)
{
  ::SendAll("UNGLINE " + Target + " [" + this->getHandle() + "]", UF_OPER,
    WATCH_KLINES, this);
  server.unkline(this->getHandle(), Target);
  Remote::Ungline(this->getHandle(), Target);
}


// help(Topic)
//
// Gets help
//
void
DCC::help(std::string Topic)
{
  StrList text = Help::getHelp(Topic);

  if (text.size() > 0)
  {
    this->send(text);
  }
  else
  {
    this->send("Help can be read at http://toast.blackened.com/oomon/");
  }
}


// getFlags()
//
// Returns a fixed length string containing the user's flags
//
std::string
DCC::getFlags() const
{
  std::string temp = "";

  if (this->isAuthed())
  {
    if (this->Flags & UF_CHANOP) temp += "C"; else temp += " ";
    if (this->Flags & UF_DLINE) temp += "D"; else temp += " ";
    if (this->Flags & UF_GLINE) temp += "G"; else temp += " ";
    if (this->Flags & UF_KLINE) temp += "K"; else temp += " ";
    if (this->Flags & UF_MASTER) temp += "M"; else temp += " ";
    if (this->Flags & UF_NICK) temp += "N"; else temp += " ";
    if (this->Flags & UF_OPER) temp += "O"; else temp += " ";
    if (this->Flags & UF_REMOTE) temp += "R"; else temp += " ";
    if (this->Flags & UF_WALLOPS) temp += "W"; else temp += " ";
    if (this->Flags & UF_CONN) temp += "X"; else temp += " ";
  }
  else
  {
    temp = "-         ";
  }
  return temp;
}


void
DCC::send(const std::string & message, const int flags,
  const WatchSet & watches)
{
  if (this->isConnected() &&
    ((flags == UF_NONE) || (flags == (this->Flags & flags))))
  {
    if (!this->watches.has(watches))
    {
#ifdef DCC_DEBUG
      cout << "DCC client not watching this type of message" << endl;
#endif
      // Missing watch...don't display status message
      return;
    }
#ifdef DCC_DEBUG
    cout << "DCC << " << message << endl;
#endif
    this->write(message + "\n");
  }
  else
  {
#ifdef DCC_DEBUG
    cout << "DCC::send(): DCC not connected or lacks required flags" << endl;
#endif
  }
}


// send(strings, flags, watches)
//
// Sends a list of strings to the DCC connection
//
void
DCC::send(StrList & messages, const int flags, const WatchSet & watches)
{
  for (StrList::iterator pos = messages.begin(); pos != messages.end(); ++pos)
  {
    this->send((*pos), flags, watches);
  }
}


void
DCC::trap(const std::string & to, std::string text)
{
  if (this->Flags & UF_MASTER)
  {
    if (to == "")
    {
      StrList output;
      TrapList::cmd(output, text, this->getHandle());
      this->send(output);
    }
    else
    {
      this->notRemote();
    }
  }
  else
  {
    this->noAccess();
  }
}


void
DCC::watch(const std::string & to, std::string text)
{
  if (this->isAuthed())
  {
    if (to == "")
    {
      if (text == "")
      {
        this->send("Watching: " + WatchSet::getWatchNames(this->watches,
	  false));
      }
      else
      {
	StrList output;

	this->watches.set(output, text);

	this->send(output);

	userConfig->setWatches(this->Handle,
	  WatchSet::getWatchNames(this->watches, true));
      }
    }
    else
    {
      this->notRemote();
    }
  }
  else
  {
    this->notAuthed();
  }
}


void
DCC::doEcho(const std::string & parm)
{
  try
  {
    if (parm == "")
    {
      this->send(std::string("*** ECHO is ") +
        (this->echoMyChatter ? "ON" : "OFF"));
    }
    else if (parm == "ON")
    {
      this->echoMyChatter = true;
      this->send("*** ECHO is now ON.");
      if (this->isAuthed())
      {
        userConfig->setEcho(this->Handle, true);
      }
    }
    else if (parm == "OFF")
    {
      this->echoMyChatter = false;
      this->send("*** ECHO is now OFF.");
      if (this->isAuthed())
      {
        userConfig->setEcho(this->Handle, false);
      }
    }
    else
    {
      this->send("*** Huh?  Turn echo ON or OFF?");
    }
  }
  catch (OOMon::botdb_error & e)
  {
    Log::Write("Error in DCC::doEcho(): " + e.why());
  }
}


void
DCC::loadConfig()
{
  try
  {
    try
    {
      this->echoMyChatter = userConfig->getEcho(this->Handle);
    }
    catch (OOMon::norecord_error)
    {
#ifdef DCC_DEBUG
      cout << "No ECHO setting for " << this->Handle << " in " <<
	userConfig->getFile() << endl;
#endif
    }

    try
    {
      StrList output;
      this->watches.set(output, userConfig->getWatches(this->Handle));
    }
    catch (OOMon::norecord_error)
    {
#ifdef DCC_DEBUG
      cout << "No WATCH setting for " << this->Handle << " in " <<
	userConfig->getFile() << endl;
#endif
    }
  }
  catch (OOMon::botdb_error & e)
  {
    Log::Write("Error while loading settings for " + this->Handle + ": " +
      e.why());
  }
}


DCC::DccCommand
DCC::getCommand(const std::string & text)
{
  if (text == "AUTH") return DCC_AUTH;
  else if (text == "QUIT") return DCC_QUIT;
  else if (text == "CHAT") return DCC_CHAT;
  else if (text == "WHO") return DCC_WHO;
  else if (text == "LINKS") return DCC_LINKS;
  else if (text == "HELP") return DCC_HELP;
  else if (text == "WATCH") return DCC_WATCH;
  else if (text == "MOTD") return DCC_MOTD;
  else if (text == "KLINE") return DCC_KLINE;
  else if (text == "UNKLINE") return DCC_UNKLINE;
  else if (text == "KCLONE") return DCC_KCLONE;
  else if (text == "KFLOOD") return DCC_KFLOOD;
  else if (text == "KBOT") return DCC_KBOT;
  else if (text == "KLINK") return DCC_KLINK;
  else if (text == "KSPAM") return DCC_KSPAM;
  else if (text == "KPERM") return DCC_KPERM;
  else if (text == "GLINE") return DCC_GLINE;
  else if (text == "UNGLINE") return DCC_UNGLINE;
  else if (text == "GUNKLINE") return DCC_UNGLINE;
  else if (text == "GCLONE") return DCC_GCLONE;
  else if (text == "GFLOOD") return DCC_GFLOOD;
  else if (text == "GBOT") return DCC_GBOT;
  else if (text == "GLINK") return DCC_GLINK;
  else if (text == "GSPAM") return DCC_GSPAM;
  else if (text == "GPERM") return DCC_GPERM;
  else if (text == "DLINE") return DCC_DLINE;
  else if (text == "KILL") return DCC_KILL;
  else if (text == "KILLLIST") return DCC_KILLLIST;
  else if (text == "KL") return DCC_KILLLIST;
  else if (text == "MULTI") return DCC_MULTI;
  else if (text == "HMULTI") return DCC_HMULTI;
  else if (text == "UMULTI") return DCC_UMULTI;
  else if (text == "VMULTI") return DCC_VMULTI;
  else if (text == "BOTS") return DCC_MULTI;
  else if (text == "VBOTS") return DCC_VMULTI;
  else if (text == "NFIND") return DCC_NFIND;
  else if (text == "TRACE") return DCC_TRACE;
  else if (text == "CLONES") return DCC_CLONES;
  else if (text == "LIST") return DCC_LIST;
  else if (text == "IPLIST") return DCC_LIST;
  else if (text == "FINDK") return DCC_FINDK;
  else if (text == "RELOAD") return DCC_RELOAD;
  else if (text == "DOMAINS") return DCC_DOMAINS;
  else if (text == "CLASS") return DCC_CLASS;
  else if (text == "DIE") return DCC_DIE;
  else if (text == "CONN") return DCC_CONN;
  else if (text == "DISCONN") return DCC_DISCONN;
  else if (text == "KTRACE") return DCC_KTRACE;
  else if (text == "KMOTD") return DCC_KMOTD;
  else if (text == "GTRACE") return DCC_GTRACE;
  else if (text == "GMOTD") return DCC_GMOTD;
  else if (text == "KINFO") return DCC_KINFO;
  else if (text == "GINFO") return DCC_GINFO;
  else if (text == "KPROXY") return DCC_KPROXY;
  else if (text == "KWINGATE") return DCC_KPROXY;
  else if (text == "GPROXY") return DCC_GPROXY;
  else if (text == "GWINGATE") return DCC_GPROXY;
  else if (text == "RAW") return DCC_RAW;
  else if (text == "OP") return DCC_OP;
  else if (text == "JOIN") return DCC_JOIN;
  else if (text == "PART") return DCC_PART;
  else if (text == "TRAP") return DCC_TRAP;
  else if (text == "UNDLINE") return DCC_UNDLINE;
  else if (text == "FINDD") return DCC_FINDD;
  else if (text == "ECHO") return DCC_ECHO;
  else if (text == "SEEDRAND") return DCC_SEEDRAND;
  else if (text == "STATUS") return DCC_STATUS;
  else if (text == "SAVE") return DCC_SAVE;
  else if (text == "LOAD") return DCC_LOAD;
  else if (text == "SET") return DCC_SET;
  else if (text == "SPAMSUB") return DCC_SPAMSUB;
  else if (text == "SPAMUNSUB") return DCC_SPAMUNSUB;
#ifdef DEBUG
  else if (text == "TEST") return DCC_TEST;
#endif
  else return DCC_UNKNOWN;
}

