// $Id: flood.cc,v 1.6 2003/02/23 23:35:51 toast Exp $

// Std C++ Headers
#include <string>

// Std C Headers
#include <time.h>

// OOMon Headers
#include "flood.h"
#include "config.h"
#include "util.h"
#include "main.h"
#include "irc.h"
#include "log.h"
#include "vars.h"
#include "userhash.h"


void
FloodList::addFlood(const std::string & nick, const std::string & userhost,
  time_t now, bool local)
{
  bool foundEntry = false;

  for (std::list<FloodEntry>::iterator pos = this->list.begin();
    pos != this->list.end(); ++pos)
  {
    if (pos->userhost == userhost)
    {
      foundEntry = true;
  
      /* if its an old old entry, let it drop to 0, then start counting
         (this should be very unlikely case)
       */
      if ((pos->last + this->getMaxTime()) < now)
      {
        pos->count = 0;
      }

      pos->count++;
      pos->last = now;
      
      if (pos->count >= this->getMaxCount())
      {
	std::string msg("Possible " + this->getType() + " flooder: " +
	  nick + " (" + userhost + ")");
        ::SendAll(msg, UF_OPER);
	Log::Write(msg);

        if (local && !Config::IsOKHost(userhost))
        {
	  doAction(nick, userhost, users.getIP(nick, userhost),
	    vars[this->getActionVar()]->getAction(),
	    vars[this->getActionVar()]->getInt(),
	    vars[this->getReasonVar()]->getString(), true);
        }
      }       
    }         
    else
    {         
      if ((pos->last + this->getMaxTime()) < now)
      {     
	this->list.erase(pos);
	--pos;
      }
    }
  }

/*
   If this is a new entry, then foundEntry will still be false
*/
  if (!foundEntry)
  {
    FloodEntry pos;

    pos.userhost = userhost;
    pos.last = now;
    pos.count = 0;

    this->list.push_back(pos);
  }
}


void
FloodList::onNotice(const std::string & notice, std::string text,
  time_t now)
{
  if (DownCase(FirstWord(text)) != "requested")
  {
    // broken.  ignore it.
    return;
  }

  if (DownCase(FirstWord(text)) != "by")
  {
    // broken.  ignore it.
    return;
  }

  std::string nick = FirstWord(text);
  std::string userhost = FirstWord(text);
  std::string serverName = FirstWord(text);

  if ((nick == "") || (userhost == ""))
  {
    // broken.  ignore it.
    return;
  }

  // Remove brackets/parentheses from userhost
  if ((userhost.length() > 0) &&
    ((userhost[0] == '[') && (userhost[userhost.length() - 1] == ']')) ||
    ((userhost[0] == '(') && (userhost[userhost.length() - 1] == ')')))
  {
    userhost = DownCase(userhost.substr(1, userhost.length() - 2));
  }

  // Remove brackets/parentheses from serverName
  if ((serverName.length() > 0) &&
    ((serverName[0] == '[') && (serverName[serverName.length() - 1] == ']')) ||
    ((serverName[0] == '(') && (serverName[serverName.length() - 1] == ')')))
  {
    serverName = DownCase(serverName.substr(1, serverName.length() - 2));
  }

  /* Don't complain about opers */
  if (Config::IsOper(userhost))
  {
    return;
  }

  ::SendAll(notice, UF_OPER);

  this->addFlood(nick, userhost, now, Same(serverName, server.getServerName()));
}

