// $Id: irc.h,v 1.25 2003/02/23 23:35:51 toast Exp $

#ifndef __IRC_H__
#define __IRC_H__

// Std C++ Headers
#include <string>

// OOMon Headers
#include "oomon.h"
#include "strtype"
#include "botsock.h"
#include "klines.h"


class IRC : public BotSock
{
public:
  IRC();

  int write(const std::string & text);

  void quit(const std::string & Message = "Normal termination");
  void notice(const std::string &, const std::string &);
  void msg(const std::string &, const std::string &);
  void ctcp(const std::string &, const std::string &);
  void ctcpReply(const std::string &, const std::string &);
  void isOn(const std::string &);
  void kline(const std::string &, const int, const std::string &,
    const std::string &);
  void unkline(const std::string &, const std::string &);
  void dline(const std::string &, const std::string &, const std::string &);
  void undline(const std::string &, const std::string &);
  void kill(const std::string &, const std::string &, const std::string &);
  void trace(const std::string & Target = "");
  void statsL(const std::string &);
  void retrace(const std::string &);
  void join(const std::string &, const std::string &);
  void part(const std::string &);
  void whois(const std::string & nick);
  void op(const std::string &, const std::string &);

  void reloadKlines(const std::string &);
  void reloadDlines(const std::string &);
  void findK(StrList &, const std::string &, bool count = false) const;
  int findAndRemoveK(const std::string & from, const std::string & userhost);
  void findD(StrList &, const std::string &, bool count = false) const;
  int findAndRemoveD(const std::string & from, const std::string & userhost);
  void subSpamTrap(const bool sub);

  std::string getServerName() const { return serverName; };

  void status(StrList & output) const;

  virtual bool onConnect();

protected:
  virtual bool onRead(std::string text);

private:
  bool amIAnOper; // duh
  bool gettingTrace;
  bool gettingKlines;
  bool gettingTempKlines;
  bool gettingDlines;
  std::string myNick;
  std::string serverName;
  KlineList klines;
  KlineList dlines;

  void onCTCP(const std::string & From, const std::string & UserHost,
    const std::string & To, std::string Text);
  void onPrivmsg(const std::string &, const std::string &,
    const std::string &, std::string);
  void onNotice(const std::string &, const std::string &,
    const std::string &, std::string);
};


extern IRC server;


#endif /* __IRC_H__ */

