// $Id: klines.cc,v 1.19 2003/05/07 21:51:13 toast Exp $

// C++ headers
#include <string>
#include <list>

#include "strtype"
#include "klines.h"
#include "irc.h"
#include "util.h"
#include "main.h"
#include "log.h"
#include "vars.h"


#ifdef DEBUG
# define KLINES_DEBUG
#endif


void KlineList::Clear()
{
  Klines.clear();
}

void KlineList::Add(const std::string & userhost, const std::string & reason)
{
  if (userhost != "")
  {
    #ifdef KLINES_DEBUG
      cout << std::string("Adding ") + lineType + "-line: " << userhost <<
	" (" << reason << ')' << endl;
    #endif
    Klines.push_back(KlineItem(userhost, reason));
  }
}


void
KlineList::find(StrList & Output, const std::string & userhost, bool count)
  const
{
  int matches = 0;
  Output.clear();
  if (!count)
  {
    Output.push_back(std::string("Finding ") + lineType + "-lines matching " +
      userhost + ':');
  }
  for (std::list<KlineItem>::const_iterator pos = Klines.begin();
    pos != Klines.end(); ++pos)
  {
    if (MatchesMask(pos->userhost, userhost))
    {
      if (!count)
      {
        Output.push_back("  " + pos->userhost + " (" + pos->reason + ')');
      }
      matches++;
    }
  }
  Output.push_back(std::string("End of FIND") + lineType + " " + userhost +
    " (" + IntToStr(Klines.size()) + " " + lineType + "-lines, " + IntToStr(matches) +
    " matches)");
}


int
KlineList::findAndRemove(const std::string & from, const std::string & userhost)
{
  int matches = 0;

  for (std::list<KlineItem>::iterator pos = Klines.begin();
    pos != Klines.end(); ++pos)
  {
    if (MatchesMask(pos->userhost, userhost))
    {
      if (lineType == 'K')
      {
        server.unkline(from, pos->userhost);
      }
      else if (lineType == 'D')
      {
        server.undline(from, pos->userhost);
      }
      matches++;
    }
  }

  return matches;
}


void KlineList::Remove(const std::string & userhost)
{
  bool GotMatch;

  do {
    GotMatch = false;
    for (std::list<KlineItem>::iterator pos = Klines.begin();
      pos != Klines.end(); ++pos)
    {
      if (Same(pos->userhost, userhost))
      {
	#ifdef KLINE_DEBUG
	  cout << std::string("Removing ") + lineType + "-line: " <<
	    pos->userhost << endl;
	#endif
	GotMatch = true;
	Klines.erase(pos);
	break;
      }
    }
  } while (GotMatch);
}

// LarzMON added K-Line for [*khknkfg@*.sprint.ca] [Clones are prohibited [Toast@LarzMON]]
void
KlineList::ParseAndAdd(std::string Text)
{
  std::string who = FirstWord(Text); // kline adder

  if (!Same("added", FirstWord(Text)))
    return;

  bool temporary = false;
  std::string time;

  if (Same("temporary", FirstWord(Text)))
  {
    temporary = true;

    time = FirstWord(Text);

    if (!Same("min.", FirstWord(Text)))
      return;

    FirstWord(Text);
  }

  if (!Same("for", FirstWord(Text)))
    return;

  // This won't work all the time, since a badly entered kline might
  // contain spaces. *shrug*
  std::string UserHost = FirstWord(Text);
  std::string Reason = Text;

  // Remove brackets
  UserHost.erase((std::string::size_type) 0, 1);
  UserHost.erase(UserHost.length() - 1, 1);
  Reason.erase((std::string::size_type) 0, 1);
  Reason.erase(Reason.length() - 1, 1);

  std::string msg;
  if (temporary)
  {
    msg = getNick(who) + " added " + time + " min. " + lineType + "-Line: " +
      UserHost + " (" + Reason + ")";
  }
  else
  {
    msg = getNick(who) + " added " + lineType + "-Line: " + UserHost + " (" +
      Reason + ")";
  }

  Log::Write(msg);
  if (lineType == 'K')
  {
    ::SendAll(msg, UF_OPER, WATCH_KLINES, NULL);
  }
  else if (lineType == 'D')
  {
    ::SendAll(msg, UF_OPER, WATCH_DLINES, NULL);
  }
  else
  {
    ::SendAll(msg, UF_OPER);
  }

  if (vars[VAR_EXTRA_KLINE_INFO]->getBool())
  {
    if (temporary)
    {
      Reason = std::string("Temporary ") + lineType + "-line " + time +
	" min. - " + Reason;
    }
    Reason = Reason + timeStamp(" (%Y/%m/%d %H.%M)");
  }

  if (!temporary || vars[VAR_TRACK_TEMP_KLINES]->getBool())
  {
    Add(UserHost, Reason);
  }
}


// Temporary K-line for [*chalupa@*.palupa] expired
void
KlineList::onExpireNotice(std::string text)
{
  if (!Same("Temporary", FirstWord(text)))
    return;
  if (!Same("K-line", FirstWord(text)))
    return;
  if (!Same("for", FirstWord(text)))
    return;

  std::string mask = FirstWord(text);
  if (mask.length() > 2)
  {
    mask = mask.substr(1, mask.length() - 2);
  }

  if (!Same("expired", FirstWord(text)))
    return;

  Remove(mask);
}


// Toast has removed the K-Line for: [*@*.monkeys.org] (1 matches)
void
KlineList::ParseAndRemove(std::string Text)
{
  std::string who = FirstWord(Text); // kline adder

  if (!Same("has", FirstWord(Text)))
    return;
  if (!Same("removed", FirstWord(Text)))
    return;
  if (!Same("the", FirstWord(Text)))
    return;

  std::string temp = FirstWord(Text);
  bool temporary = false;

  if (Same("temporary", temp))
  {
    temporary = true;

    if (!Same("K-Line", FirstWord(Text)))
      return;
  }
  else if (!Same("K-Line", temp))
    return;

  if (!Same("for:", FirstWord(Text)))
    return;

  // Again, spaces in the kline mask will screw this up.
  std::string UserHost = FirstWord(Text);

  // Remove brackets
  UserHost.erase((std::string::size_type) 0, 1);
  UserHost.erase(UserHost.length() - 1, 1);

  std::string msg;
  if (temporary)
  {
    msg = getNick(who) + " removed temp. " + lineType + "-Line: " +
      UserHost;
  }
  else
  {
    msg = getNick(who) + " removed " + lineType + "-Line: " + UserHost;
  }

  Log::Write(msg);
  if (lineType == 'K')
  {
    ::SendAll(msg, UF_OPER, WATCH_KLINES, NULL);
  }
  else if (lineType == 'D')
  {
    ::SendAll(msg, UF_OPER, WATCH_DLINES, NULL);
  }
  else
  {
    ::SendAll(msg, UF_OPER);
  }

  if (!temporary || vars[VAR_TRACK_TEMP_KLINES]->getBool())
  {
    Remove(UserHost);
  }
}

