// $Id: links.cc,v 1.9 2003/01/22 20:24:30 toast Exp $

// C++ headers
#include <string>
#include <list>

#include "oomon.h"
#include "links.h"
#include "util.h"


#ifdef DEBUG
# define LINKS_DEBUG
#endif


Links::Links()
{
  Name = "";
}

Links::Links(const std::string & name)
{
  #ifdef LINKS_DEBUG
    cout << "Links(" << name << ')' << endl;
  #endif
  Name = name;
}

Links::~Links()
{
  Clear();
}

void Links::Clear()
{
  Name = "";
  Children.clear();
}

bool Links::Exists(const std::string & name)
{
  if (Same(Name, name))
  {
    return true;
  }
  for (std::list<Links>::iterator pos = Children.begin();
    pos != Children.end(); ++pos)
  {
    if (pos->Exists(name))
    {
      return true;
    }
  }
  return false;
}

bool Links::Link(const std::string & parent, const std::string & name)
{
  if (Same(Name, parent))
  {
    #ifdef LINKS_DEBUG
      cout << "Linking " << name << " to " << Name << endl;
    #endif
    Children.push_back(Links(name));
    return true;
  }
  else
  {
    for (std::list<Links>::iterator pos = Children.begin();
      pos != Children.end(); ++pos)
    {
      if (pos->Link(parent, name))
      {
        return true;
      }
    }
  }
  return false;
}

bool Links::Unlink(const std::string & name)
{
  for (std::list<Links>::iterator pos = Children.begin();
    pos != Children.end(); ++pos)
  {
    if (Same(pos->GetName(), name))
    {
      #ifdef LINKS_DEBUG
	cout << "Unlinking " << pos->GetName() << " from " << Name << endl;
      #endif
      Children.erase(pos);
      return true;
    }
  }
  for (std::list<Links>::iterator pos = Children.begin();
    pos != Children.end(); ++pos)
  {
    if (pos->Unlink(name))
    {
      return true;
    }
  }
  return false;
}

void Links::SetName(const std::string & name)
{
  #ifdef LINKS_DEBUG
    cout << "Changing link name to " << name << endl;
  #endif
  Name = name;
}

void Links::GetTree(StrList & list, const std::string & parent)
{
  if (Name == "")
  {
    return;
  }
  list.push_back(parent + " " + Name);
  for (std::list<Links>::iterator pos = Children.begin();
    pos != Children.end(); ++pos)
  {
    pos->GetTree(list, Name);
  }
}

void Links::GetLinks(StrList & Output,
  const std::string & prefix)
{
  Output.push_back(prefix + Name);
  if (Children.size() > 0)
  {
    int i = 0;
    for (std::list<Links>::iterator pos = Children.begin();
      pos != Children.end(); ++pos)
    {
      if (i < Children.size() - 1)
      {
	Output.push_back(prefix + "|\\");
	pos->GetLinks(Output, prefix + "| ");
      }
      else
      {
	Output.push_back(prefix + " \\");
	pos->GetLinks(Output, prefix + "  ");
      }
      i++;
    }
  }
}

