// $Id: proxy.h,v 1.18 2003/02/14 22:26:45 toast Exp $

#ifndef __PROXY_H__
#define __PROXY_H__

// Std C++ headers
#include <string>
#include <list>

// Std C headers
#include <sys/types.h>
#include <unistd.h>

// OOMon headers
#include "strtype"
#include "oomon.h"
#include "botsock.h"


class Proxy : public BotSock
{
public:
  enum ProxyType { SOCKS4, WINGATE, HTTP };

  Proxy(const std::string & hostname, const std::string & nick,
    const std::string & userhost);
  virtual ~Proxy() { };

  static void check(const std::string &, const std::string &,
    const std::string &, const std::string &);
  static bool connect(Proxy *newProxy, const std::string & address,
    const BotSock::Port port);
  static void processAll(const fd_set & readset, const fd_set & writeset);
  static void setAllFD(fd_set & readset, fd_set & writeset);
  static bool isChecking(const std::string &, const BotSock::Port, ProxyType);
  static bool isSafe(const std::string & address, const BotSock::Port port,
    const ProxyType type);
  static void addSafeHost(const std::string & address, const BotSock::Port port,
    const ProxyType type);

  static void status(StrList & output);

protected:
  void detectedProxy();

  std::string getAddress() const { return this->address; };
  BotSock::Port getPort() const { return this->port; };

  std::string getHostname() const { return this->hostname; };
  std::string getNick() const { return this->nick; };
  std::string getUserhost() const { return this->userhost; };

  virtual ProxyType getType() const = 0;
  virtual std::string getTypeName() const = 0;

private:
  typedef std::list<Proxy *> ProxyList;
  static ProxyList items;

  class CacheEntry
  {
  public:
    CacheEntry() { this->ip = INADDR_NONE; };
    bool isEmpty() const { return (INADDR_NONE == this->ip); };
    BotSock::Address ip;
    BotSock::Port port;
    ProxyType type;
    time_t checked;
  };
  typedef std::vector<CacheEntry> Cache;
  static Cache safeHosts;
  static const time_t CACHE_EXPIRE;

  std::string address, hostname, nick, userhost;
  BotSock::Port port;

  bool connect(const std::string & address, const BotSock::Port port);
};

#endif /* __PROXY_H__ */

