// $Id: remote.cc,v 1.29 2003/05/28 15:56:50 toast Exp $

// Std C++ headers
#include <string>
#include <vector>
#include <list>

// Std C headers
#include <sys/errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <unistd.h>
#include <time.h>
#include <signal.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>

// OOMon Headers
#include "strtype"
#include "oomon.h"
#include "remote.h"
#include "config.h"
#include "log.h"
#include "irc.h"
#include "dcclist.h"
#include "util.h"
#include "main.h"


#ifdef DEBUG
# define REMOTE_DEBUG
#endif


int Remote::count = 0;
bool Remote::listening = false;
int Remote::remote_socket = 0;
class Remote* Remote::items[MAX_REMOTES];

void Remote::SetFD(fd_set & theset)
{
  if (listening)
    FD_SET(remote_socket, &theset);
  for (int i = 0; i < count; i++)
    items[i]->PrivSetFD(theset);
}

int Remote::Process(const fd_set & theset)
{
  if (FD_ISSET(remote_socket, &theset))
    Connect();
  for (int i = 0; i < count; i++)
    if (items[i]->PrivProcess(theset)) {
      delete items[i];
      for (int j = i; j < count-1; j++)
        items[j] = items[j+1];
      count--;
    }
  return 0;
}

void Remote::Send(const std::string & From, const std::string & Command,
        const std::string & To, const std::string & Params, const class Remote *exception)
{
  std::string ToBot;
  if (To.find('@') == std::string::npos)
    ToBot = To;
  else
    ToBot = GetBotHB(To);
  for (int i = 0; i < count; i++)
    if (items[i] != exception)
      if ((ToBot == "*") || items[i]->PrivFindBot(ToBot))
        items[i]->PrivSend(From, Command, To, Params);
}

void Remote::Chat(const std::string & From, const std::string & To,
        const std::string & Text, const class Remote *exception)
{
  Send(From, REM_CHAT, To, ":" + Text, exception);
}

void Remote::Die(const std::string & From, const std::string & To,
        const class Remote *exception)
{
  Send(From + "@" + Config::GetNick(), REM_DIE, To, "", exception);
}

void Remote::Kline(const std::string & From, const std::string & To,
        const int Minutes, const std::string & Target, const std::string & Reason)
{ 
  Send(From + "@" + Config::GetNick(), REM_KLINE, To, IntToStr(Minutes) +
    " " + Target + " :" + Reason);
}

void Remote::KClone(const std::string & From, const std::string & To,
        const int Minutes, const std::string & Target)
{   
  Send(From + "@" + Config::GetNick(), REM_KCLONE, To, IntToStr(Minutes) +
    " " + Target);
}     
      
void Remote::KFlood(const std::string & From, const std::string & To,
        const int Minutes, const std::string & Target)
{   
  Send(From + "@" + Config::GetNick(), REM_KFLOOD, To, IntToStr(Minutes) +
    " " + Target);
}

void Remote::KBot(const std::string & From, const std::string & To,
        const int Minutes, const std::string & Target)
{ 
  Send(From + "@" + Config::GetNick(), REM_KBOT, To, IntToStr(Minutes) +
    " " + Target);
} 
    
void Remote::KSpam(const std::string & From, const std::string & To,
        const int Minutes, const std::string & Target)
{     
  Send(From + "@" + Config::GetNick(), REM_KSPAM, To, IntToStr(Minutes) +
    " " + Target);
}

void Remote::KLink(const std::string & From, const std::string & To,
        const int Minutes, const std::string & Target)
{ 
  Send(From + "@" + Config::GetNick(), REM_KLINK, To, IntToStr(Minutes) +
    " " + Target);
}
        
void Remote::KTrace(const std::string & From, const std::string & To,
        const int Minutes, const std::string & Target)
{
  Send(From + "@" + Config::GetNick(), REM_KTRACE, To, IntToStr(Minutes) +
    " " + Target);
}

void Remote::KMotd(const std::string & From, const std::string & To,
        const int Minutes, const std::string & Target)
{
  Send(From + "@" + Config::GetNick(), REM_KMOTD, To, IntToStr(Minutes) +
    " " + Target);
}       
 
void Remote::KInfo(const std::string & From, const std::string & To,
        const int Minutes, const std::string & Target)
{
  Send(From + "@" + Config::GetNick(), REM_KINFO, To, IntToStr(Minutes) +
    " " + Target);
}       
 
void Remote::KProxy(const std::string & From, const std::string & To,
        const int Minutes, const std::string & Target)
{
  Send(From + "@" + Config::GetNick(), REM_KPROXY, To, IntToStr(Minutes) +
    " " + Target);
}       
 
void Remote::KPerm(const std::string & From, const std::string & To,
        const std::string & Target, const std::string & Reason)
{
  Send(From + "@" + Config::GetNick(), REM_KPERM, To, Target +
    " :" + Reason);
}       
 
void Remote::Unkline(const std::string & From, const std::string & To,
        const std::string & Target)
{
  Send(From + "@" + Config::GetNick(), REM_UNKLINE, To, Target);
}
        
void Remote::Status(const std::string & From, const std::string & To,
        const std::string & Text, const class Remote *exception)
{   
  Send(From, REM_STATUS, To, ":" + Text, exception);
}

void Remote::Status(const std::string & To, const std::string & Text,
        const class Remote *exception)
{ 
  Send(Config::GetNick(), REM_STATUS, To, ":" + Text, exception);
}

void Remote::Dline(const std::string & From, const std::string & To, 
        const std::string & Target, const std::string & Reason)
{
  Send(From + "@" + Config::GetNick(), REM_DLINE, To, Target + " :" +
    Reason);
}

void Remote::Undline(const std::string & From, const std::string & To,
        const std::string & Target)
{
  Send(From + "@" + Config::GetNick(), REM_UNDLINE, To, Target);
}
        
void Remote::Conn(const std::string & From, const std::string & To, 
        const std::string & Target)
{
  if ((To == "") || Same(To, Config::GetNick())) {
    Connect(Target);
  } else
    Send(From + "@" + Config::GetNick(), REM_CONNECT, To, Target);
}
        
void Remote::Disconn(const std::string & From, const std::string & To,
        const std::string & Target)
{   
  if ((To == "") || Same(To, Config::GetNick())) {
  } else
    Send(From + "@" + Config::GetNick(), REM_DISCONN, To, Target);
}       
 
void Remote::Who(const std::string & From, const std::string & To)  
{   
  class Remote *temp = FindBot(To);
  if (temp != NULL)
    temp->PrivSend(From + "@" + Config::GetNick(), REM_WHO, To); 
}       
 
void Remote::GetLinks(StrList & Output)
{
  Output.clear();
  Output.push_back(Config::GetNick());
  if (count > 0)
  {
    for (int i = 0; i < count; i++)
    {
      if (i < count-1)
      {
        Output.push_back("|\\");
        items[i]->PrivGetLinks(Output, "| ");
      }
      else
      {
        Output.push_back(" \\");
        items[i]->PrivGetLinks(Output, "  ");
      }
    }
  }
}

void Remote::DoLinks(const std::string & From, const std::string & To)
{       
  class Remote *temp = FindBot(To);
  if (temp != NULL) 
    temp->PrivSend(From + "@" + Config::GetNick(), REM_LINKS, To);
}

void Remote::Gline(const std::string & From, const int Minutes,
        const std::string & Target, const std::string & Reason)
{
  Send(From + "@" + Config::GetNick(), REM_GLINE, "*", IntToStr(Minutes) +
    " " + Target + " :" + Reason);
} 
    
void Remote::Ungline(const std::string & From, const std::string & Target)
{
  Send(From + "@" + Config::GetNick(), REM_UNGLINE, "*", Target);
}       
 
void Remote::MOTD(const std::string & From, const std::string & To)
{ 
  Send(From + "@" + Config::GetNick(), REM_MOTD, To, ""); 
}

int Remote::Listen()
{
  struct sockaddr_in socketname;
  BotSock::Port port = Config::GetPort();

  if (listening)
    close(remote_socket);

  memset((char *)&socketname, 0, sizeof(struct sockaddr));
  socketname.sin_family = AF_INET;
  socketname.sin_addr.s_addr = INADDR_ANY;
  socketname.sin_port = htons(port);

  if ((remote_socket = socket(PF_INET, SOCK_STREAM, 6)) < 0) {
    cerr << "Can't create socket for Remote::Listen()" << endl;
    return INVALID;
  }

  int optval = 1;
  setsockopt(remote_socket, SOL_SOCKET, SO_REUSEADDR, (char *) &optval,
    sizeof(optval));
  setsockopt(remote_socket, SOL_SOCKET, SO_KEEPALIVE, (char *) &optval,
    sizeof(optval));

  if (bind(remote_socket, (struct sockaddr *)&socketname,
    sizeof(socketname)) < 0) {
    cerr << "Can't bind port " << port << endl;
    return INVALID;
  }

  if (listen(remote_socket, 4) < 0) {
    cerr << "Can't listen on remote port" << endl;
    return INVALID;
  }
  Log::Write("Listening to port " + IntToStr(port) + " for linking bots");
  #ifdef REMOTELIST_DEBUG
    cout << "Listening to port " << port << " for linking bots" << endl;
  #endif
  listening = true;
}

int Remote::Connect()
{
  class Remote *temp;

  temp = new Remote;

  if (!temp->PrivConnect(remote_socket)) {
    items[count++] = temp;
    return 0;
  } else
    delete temp;
  return INVALID;
}

int Remote::Connect(const std::string & BotHandle)
{ 
  class Remote *temp;

  temp = new Remote;

  if (!temp->PrivConnect(BotHandle)) {
    items[count++] = temp;
  } else
    delete temp;
  return INVALID;
} 
  
void Remote::GetTree(StrList & list, class Remote *exception)
{ 
  list.clear();
  for (int i = 0; i < count; i++)
  {
    if (items[i] != exception)
    {
      items[i]->GetTree(list, Config::GetNick());
    }
  }
} 
    
bool Remote::IsLinked(const std::string & Name) 
{   
  for (int i = 0; i < count; i++)
    if (Same(Name, items[i]->GetName()))
      return true;
  return false;
}   
   
class Remote *Remote::FindBot(const std::string & To)
{ 
  for (int i = 0; i < count; i++)
    if (items[i]->FindBot(To))
      return items[i];
  return NULL;
}

bool Remote::Pass(const std::string & From, const std::string & Command,
        const std::string & To, const std::string & Text)
{
  if (To == "*") {
    for (int i = 0; i < count; i++)
      items[i]->PrivSend(From, Command, To, Text);
  } else {
    class Remote *temp;
    if (To.find('@') != std::string::npos)
      temp = FindBot(GetBotHB(To));
    else
      temp = FindBot(To);
    if (temp != NULL)
      temp->PrivSend(From, Command, To, Text);
    else
      return false;
  }
  return true;
}

Remote::Remote()
{
  plug = 0;
  connected = false;
  Host = "";
  Handle = "";
  Flags = 0;
  authorized = false;
  ReadyToRock = false;
}

Remote::~Remote()
{
  if (connected)
    if (authorized)
      Status("Lost connection to " + Handle);
    else
      Status("Lost bot connection before registration");
  Disconnect();
}

int Remote::Write(const std::string & Text)
{
  if (connected) {
    #ifdef REMOTE_DEBUG
      cout << "Remote << " << Text;
    #endif
    send(plug, Text.c_str(), Text.length(), 0);
  } else
    return INVALID;
  return 0;
}

int Remote::PrivSend(const std::string & From, const std::string & Command,
	const std::string & To, const std::string & Params)
{
  if (authorized && ReadyToRock)
    return Write(":" + From + " " + Command + " " + To + " " + Params + '\n');
  return 0;
}

void Remote::PrivSetFD(fd_set & theset)
{
  if (connected)
    FD_SET(plug, &theset);
}

int Remote::PrivProcess(const fd_set & theset)
{
  if (connected)
    if (FD_ISSET(plug, &theset)) {
      std::string text;
      if (int i = ReadLine(plug, text) <= 0) 
        return INVALID;
      if (text != "") {
        LastActive = time((time_t *)NULL);
        #ifdef REMOTE_DEBUG
          cout << "Remote >> " << text << endl;
        #endif 
        return Parse(text);
      } else
	return 0;
    } else
      return 0;
  return INVALID;
}

int Remote::PrivConnect(const int remote_socket)
{
  struct sockaddr_in incoming_addr; 
  struct hostent *host_seen;
  socklen_t addrlen;
  int lnth;     
  char host[MAX_BUFF];

  addrlen = sizeof(struct sockaddr);
  if((plug = accept(remote_socket,
                     (struct sockaddr *)&incoming_addr, &addrlen)) < 0 )
    {
      cerr << "Error in remote tcm connect on accept" << endl;
      return INVALID;
    }
  LastActive = time((time_t *)NULL);

  connected = true;
 
  host_seen = gethostbyaddr((char *)&incoming_addr.sin_addr.s_addr,
                            4, AF_INET);

  if(host_seen)
    strncpy(host, (char *)host_seen->h_name, MAX_HOST-1);
  else
    strncpy(host, inet_ntoa(incoming_addr.sin_addr), MAX_HOST-1);

  Host = host;

  if (!Config::IsLinkable(Host)) {
    #ifdef REMOTE_DEBUG
      cout << "Bot connection refused for " << Host << " (No L: line)" << endl;
    #endif
    Status("Bot connection refused for " + Host + " (No L: line)");
    Log::Write("Bot connection refused for " + Host + " (No L: line)");
    Disconnect();
    return INVALID;
  }

  #ifdef REMOTE_DEBUG
    cout << "Bot connectiong from " << Host << endl;
  #endif

  Status("Bot connection from " + Host);
  Log::Write("Bot connection from " + Host);

  authorized = false;

  return 0;
}

void Remote::Disconnect()
{
  if (connected) {
    #ifdef REMOTE_DEBUG
      cout << "Closing bot connection from " << Host << endl;
    #endif
    close(plug);
  }
  plug = 0;
  connected = false;
  ReadyToRock = false;
  ServerTree.Clear();
}

void Remote::Status(const std::string & Text)
{
  //DCC::Status(Text);
  Remote::Status("*", Text, this);
}

int Remote::Parse(std::string Text)
{
  std::string FromBot;
  if (Text == "")
    return 0;
  if (authorized) {
    StrVector Params;
    std::string From, Command, To;
    if (Text[0] == ':') {
      Text.erase((std::string::size_type) 0, 1);
      From = FirstWord(Text);
      Command = FirstWord(Text);
      To = FirstWord(Text);
    } else {
      From = Handle;
      Command = FirstWord(Text);
      To = FirstWord(Text);
    }
    SplitIRC(Params, Text);
    if (From.find('@') != std::string::npos)
      FromBot = GetBotHB(From);
    else
      FromBot = From;
    if (!ServerTree.Exists(FromBot)) {
      Write("ERROR * :Wrong direction for message from " + FromBot + '\n');
      Status("Closing link to " + Handle +
	" (Wrong direction for message from " + FromBot + ')');
      Disconnect();
      return INVALID;
    }
    int commandnum;
    if ((commandnum = atoi(Command.c_str())) != 0)
      switch (commandnum) {
      case 001:
	if (!ReadyToRock) {
	  StrList tree;
	  Remote::GetTree(tree, this);
	  for (StrList::iterator pos = tree.begin();
	    pos != tree.end(); ++pos)
	  {
            Write(":" + Config::GetNick() + " 009 " + Handle + " " + *pos +
	      '\n');
	  }
	  Write(":" + Config::GetNick() + " 010 * :Ready\n");
	}
	break;
      case 9:
	{
	  std::string parent = Params[0];
	  std::string name = Params[1];
	  if (!ReadyToRock)
	    Remote::Pass(From, "009", "*", parent + " " + name);
	  ServerTree.Link(parent, name);
	}
	break;
      case 10:
	if (!ReadyToRock) {
	  ReadyToRock = true;
	  Status("Bot " + Handle + " has registered");
	}
      }
    else {
      if (ReadyToRock)
        if ((To != "*") && (!Same(To, Config::GetNick())) &&
	  (!Same(GetBotHB(To), Config::GetNick()))) {
          if (!Remote::Pass(From, Command, To, Text)) {
	    Write("ERROR * :Target not found: " + To + '\n');
	    Status("Closing link to " + Handle + " (Target not found: " +
	      To + ')');
	    Disconnect();
	    return INVALID;
          }
          return 0;
        }
      switch (RemoteCommand(Command)) {
      case RC_ERROR:
        Status("Closing link to " + Handle + " (" + Params[0] + ')');
	Disconnect();
        return INVALID;
      case RC_STATUS:
	if ((To == "*") || (!Same(GetBotHB(To), Config::GetNick())))
	  Remote::Status(From, To, Params[0], this);
	//if ((To == "*") || (Same(GetBotHB(To), Config::GetNick())))
	  //DCC::Status(From, GetHandleHB(To), Params[0]);
	break;
      case RC_CHAT:
	if ((To == "*") || (!Same(GetBotHB(To), Config::GetNick())))
	  Remote::Chat(From, To, Params[0], this);
	if ((To == "*") || (Same(GetBotHB(To), Config::GetNick())))
	  if (!clients.chat(From, GetHandleHB(To), Params[0]))
	    Send(Config::GetNick(), REM_STATUS, From, ":User not found: " + To);
	break;
      case RC_WHO:
	if (Same(To, Config::GetNick())) {
	  StrList Output;
	  clients.who(Output);
	  for (StrList::iterator pos = Output.begin();
	    pos != Output.end(); ++pos)
	  {
	    Send(Config::GetNick(), REM_STATUS, From, ":" + *pos);
	  }
	}
	break;
      case RC_KLINE:
	if (Same(To, Config::GetNick())) {
	  int Minutes = atoi(Params[0].c_str());
	  std::string Target = Params[1];
	  std::string Reason = Params[2];
	  server.kline(From, Minutes, Target, Reason);
	}
	break;
      case RC_LINKS:
	if (Same(To, Config::GetNick())) {
	  StrList Output;
	  Remote::GetLinks(Output);
	  for (StrList::iterator pos = Output.begin();
	    pos != Output.end(); ++pos)
	  {
	    Send(Config::GetNick(), REM_STATUS, From, ":" + *pos);
	  }
	}
	break;
      case RC_MOTD:
	if (Same(To, Config::GetNick())) {
	  StrList Output;
	  ::motd(Output);
	  for (StrList::iterator pos = Output.begin();
	    pos != Output.end(); ++pos)
	  {
	    Send(Config::GetNick(), REM_STATUS, From, ":" + *pos);
	  }
	}
	break;
      case RC_DIE:
	if (Same(To, Config::GetNick())) {
	  Status(From + " says I have to die now...");
	  Log::Write("DIE requested by " + From);
	  gracefuldie(SIGTERM); 
	}
	break;
      case RC_UNKNOWN:
        Write("ERROR " + Handle + " :Unknown command: " + Command + '\n');
	Status("Closing link to " + Handle + " (Unknown command: " +
	  Command + ')');
	Disconnect();
        return INVALID;
        break;
      }
    }
  } else {
    std::string Version = FirstWord(Text);
    if (Version != "OOMon") {
      Write("ERROR * :Incompatible Version\n");
      Status("Rejecting link (Incompatible Version)");
      Disconnect();
      return INVALID;
    }
    std::string BotHandle = FirstWord(Text);
    if (Remote::IsLinked(BotHandle)) {
      Write("ERROR * :" + BotHandle + " already exists!\n");
      Status("Rejecting link (" + BotHandle + " already exists!)");
      Disconnect();
      return INVALID;
    }
    if (Config::AuthBot(BotHandle, Host, Text)) {
      Handle = BotHandle;
      ServerTree.SetName(Handle);
      Write(":" + Config::GetNick() + " 001 * :Welcome to OOMon-Net\n");
      StrList tree;
      Remote::GetTree(tree, this);
      for (StrList::iterator pos = tree.begin();
	pos != tree.end(); ++pos)
      {
        Write(":" + Config::GetNick() + " 009 * " + *pos + '\n');
      }
      Write(":" + Config::GetNick() + " 010 * :Ready\n");
      ReadyToRock = false;
      authorized = true;
    } else {
      Write("ERROR * :Access Denied\n");
      Status("Rejecting link (Access Denied)");
      Disconnect();
      return INVALID;
    }
  }
  return 0;
}

int Remote::PrivConnect(const std::string & BotHandle)
{
  struct sockaddr_in	socketname;
  struct sockaddr_in	localaddr;
  struct hostent	*remote_host;
  struct hostent	*local_host;
  char			buf[MAX_BUFF];
  char			s[MAX_BUFF];
  int			portnum = 6667;
  char			server[MAX_HOST],
			*hold;
  unsigned long		remoteaddr;
  std::string		Host;
  BotSock::Port		port;
  std::string		Passwd;

  #ifdef REMOTE_DEBUG
    cout << "Connecting to remote bot: " << BotHandle << endl;
  #endif

  Log::Write("Connecting to remote bot: " + BotHandle);

  Handle = BotHandle;
  ServerTree.SetName(Handle);

  Config::GetConn(Handle, Host, port, Passwd, Flags);

  if ((plug = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
    cerr << "Error: Can't assign fd for socket!" << endl;
    return INVALID;
  }

  int optval = 1;
  setsockopt(plug, SOL_SOCKET, SO_REUSEADDR, (char *) &optval, sizeof(optval));
  setsockopt(plug, SOL_SOCKET, SO_KEEPALIVE, (char *) &optval, sizeof(optval));

  (void)memset((char *)&socketname, 0, sizeof(socketname));

  (void)memset((char *)&localaddr, 0, sizeof(struct sockaddr_in));
  if (Config::GetHostName() != "")
    if (local_host = gethostbyname(Config::GetHostName().c_str()))
      (void)memcpy((char *) local_host->h_addr,
	(char *) &localaddr.sin_addr,
	local_host->h_length);
  localaddr.sin_family = AF_INET;
  localaddr.sin_port = 0;

  if (bind(plug, (struct sockaddr *) &localaddr, sizeof(socketname)) < 0)
    cerr << "Error: Unable to bind virtual host" << endl;

  if ((remote_host = gethostbyname(Host.c_str())) == (struct hostent *) 0) {
    cerr << "Error: Unknown host: " << Host << endl;
    return INVALID;
  }
  (void)memcpy((char *) remote_host->h_addr,
	(char *) &socketname.sin_addr,
	remote_host->h_length);
  socketname.sin_family = AF_INET;
  socketname.sin_port = htons(port);

  if (connect(plug, (struct sockaddr *) &socketname, sizeof(socketname)) < 0)
    if (errno != EINTR) {
      cerr << "Error: Connect " << errno << " (" << strerror(errno) << ")" << endl;
      return INVALID;
    }

  connected = true;

  #ifdef REMOTE_DEBUG
    cout << "Connected to bot " << BotHandle << endl;
  #endif
  Log::Write("Connected to bot " + BotHandle);
  Status("Connected to bot " + BotHandle);

  authorized = true;
  ReadyToRock = false;

  Write("OOMon " + Config::GetNick() + " " + Passwd + '\n');

  return 0;
}

void Remote::GetTree(StrList & list, const std::string & parent)
{
  ServerTree.GetTree(list, parent);
}

void Remote::PrivGetLinks(StrList & Output,
  const std::string & prefix)
{
  ServerTree.GetLinks(Output, prefix);
}

bool Remote::PrivFindBot(const std::string & Name)
{
  return ServerTree.Exists(Name);
}
