// $Id: remote.h,v 1.13 2003/01/20 20:49:30 toast Exp $

#ifndef _REMOTE_H_
#define _REMOTE_H_

#include <string>
#include <list>

#include "strtype"
#include "oomon.h"
#include "links.h"

class Remote
{
private:
  // Class Variables
  static int remote_socket;
  static bool listening;
  static class Remote *items[];
  static int count;

  // Class Functions
  static void Send(const std::string &, const std::string &,
    const std::string &, const std::string &,
    const class Remote *exception = 0);
  static int Connect();
  static int Connect(const std::string &);

  // Instance Variables
  int plug;
  bool connected, authorized;
  bool ReadyToRock;
  time_t LastActive;
  std::string Host;
  std::string Handle;
  int Flags;
  Links ServerTree;

  // Instance Functions
  int  Write(const std::string &);
  int  PrivConnect(int);
  int  PrivConnect(const std::string &);
  int  PrivProcess(const fd_set &);
  void PrivSetFD(fd_set &);
  bool PrivFindBot(const std::string &);
  void PrivGetLinks(StrList &, const std::string &);
  int  PrivSend(const std::string & From, const std::string & Command,
    const std::string & To, const std::string & Params = "");
public:
  // Class Functions
  static void SetFD(fd_set & theset);
  static int  Process(const fd_set & theset);
  static int  Listen();
  static void Chat(const std::string & From, const std::string & To,
    const std::string & Text,
    const class Remote *exception = (class Remote *) 0);
  static void Die(const std::string & From, const std::string & To,
        const class Remote *exception = (class Remote *) 0);
  static void Kline(const std::string &, const std::string &, const int,
    const std::string &, const std::string &);
  static void KClone(const std::string &, const std::string &, const int,
    const std::string &);
  static void KFlood(const std::string &, const std::string &, const int,
    const std::string &);
  static void KBot(const std::string &, const std::string &, const int,
    const std::string &);
  static void KSpam(const std::string &, const std::string &, const int,
    const std::string &);
  static void KLink(const std::string &, const std::string &, const int,
    const std::string &);
  static void KInfo(const std::string &, const std::string &, const int,
    const std::string &);
  static void KWingate(const std::string &, const std::string &, const int,
    const std::string &);
  static void KProxy(const std::string &, const std::string &, const int,
    const std::string &);
  static void KTrace(const std::string &, const std::string &, const int,
    const std::string &);
  static void KMotd(const std::string &, const std::string &, const int,
    const std::string &);
  static void KPerm(const std::string &, const std::string &,
    const std::string &, const std::string &);
  static void Unkline(const std::string &, const std::string &,
    const std::string &);
  static void Status(const std::string &, const std::string &,
    const class Remote *exception = (class Remote *) 0);
  static void Status(const std::string &, const std::string &,
    const std::string &, const class Remote *exception = (class Remote *) 0);
  static void Dline(const std::string &, const std::string &,
    const std::string &, const std::string &);
  static void Undline(const std::string &, const std::string &,
    const std::string &);
  static void Conn(const std::string & From, const std::string & To,
    const std::string & Target);
  static void Disconn(const std::string & From, const std::string & To,
    const std::string & Target);
  static void Who(const std::string & From, const std::string & To);
  static void GetLinks(StrList & Output);
  static void DoLinks(const std::string &, const std::string &);
  static void Gline(const std::string & From, const int Minutes,
    const std::string & Target, const std::string & Reason);
  static void Ungline(const std::string &, const std::string &);
  static void MOTD(const std::string &, const std::string &);
  static void GetTree(StrList &, class Remote *exception);
  static bool IsLinked(const std::string &);
  static class Remote *FindBot(const std::string &);
  static bool Pass(const std::string &, const std::string &,
    const std::string &, const std::string &);

  // Instance Functions
  Remote();
  ~Remote();
  void Disconnect();
  void Status(const std::string &);
  int  Parse(std::string);
  std::string GetName() const { return Handle; }
  void GetTree(StrList &, const std::string &);
};

#define REM_NOP		"NOP"
#define REM_CHAT	"CHAT"
#define REM_DIE		"DIE"
#define REM_CONNECT	"CONNECT"
#define REM_DISCONN	"DISCONN"
#define REM_STATUS	"STATUS"
#define REM_CHAT	"CHAT"
#define REM_WHO		"WHO"
#define REM_LINKS	"LINKS"
#define REM_KLINE	"KLINE"
#define REM_KCLONE	"KCLONE"
#define REM_KFLOOD	"KFLOOD"
#define REM_KBOT	"KBOT"
#define REM_KSPAM	"KSPAM"
#define REM_KLINK	"KLINK"
#define REM_KTRACE	"KTRACE"
#define REM_KMOTD	"KMOTD"
#define REM_KINFO	"KINFO"
#define REM_KPROXY	"KPROXY"
#define REM_KPERM	"KPERM"
#define REM_UNKLINE	"UNKLINE"
#define REM_GLINE	"GLINE"
#define REM_GCLONE	"GCLONE"
#define REM_GFLOOD	"GFLOOD"
#define REM_GBOT	"GBOT"
#define REM_GSPAM	"GSPAM"
#define REM_GLINK	"GLINK"
#define REM_GTRACE	"GTRACE"
#define REM_GMOTD	"GMOTD"
#define REM_GINFO	"GINFO"
#define REM_GPROXY	"GPROXY"
#define REM_GPERM	"GPERM"
#define REM_UNGLINE	"UNGLINE"
#define REM_DLINE	"DLINE"
#define REM_UNDLINE	"UNDLINE"
#define REM_MOTD	"MOTD"

#define REM_WELCOME	"001"
#define REM_INITLINK	"009"
#define REM_INITEND	"010"

#define RC_UNKNOWN	0
#define RC_CONNECT	1
#define RC_DISCONN	2
#define RC_STATUS	3
#define RC_CHAT		4
#define RC_WHO		5
#define RC_LINKS	6
#define RC_KLINE	7
#define RC_KCLONE	8
#define RC_KFLOOD	9
#define RC_KBOT		10
#define RC_KSPAM	11
#define RC_KLINK	12
#define RC_KPERM	13
#define RC_UNKLINE	14
#define RC_GLINE	15
#define RC_GCLONE	16
#define RC_GFLOOD	17
#define RC_GBOT		18
#define RC_GSPAM	19
#define RC_GLINK	20
#define RC_GPERM	21
#define RC_UNGLINE	22
#define RC_DLINE	23
#define RC_DIE		24
#define RC_MOTD		25
#define RC_KTRACE	26
#define RC_KMOTD	27
#define RC_GTRACE	28
#define RC_GMOTD	29
#define RC_KINFO	30
#define RC_GINFO	31
#define RC_KPROXY	32
#define RC_GPROXY	33
#define RC_UNDLINE	34

#define RC_ERROR	500

#endif
