// $Id: trap.h,v 1.8 2003/02/20 19:06:57 toast Exp $

#ifndef __TRAP_H__
#define __TRAP_H__

// Std C++ Headers
#include <fstream>
#include <string>
#include <list>

// OOMon Headers
#include "strtype"
#include "botsock.h"


enum TrapAction
{
  TRAP_ECHO, TRAP_KILL, TRAP_KLINE, TRAP_KLINE_HOST, TRAP_KLINE_DOMAIN,
  TRAP_KLINE_IP, TRAP_KLINE_USERNET, TRAP_KLINE_NET, TRAP_DLINE_IP,
  TRAP_DLINE_NET
};


class Trap
{
public:
  Trap(const TrapAction action, const long timeout, const std::string & nick,
    const std::string & userhost, const std::string & reason);

  bool matches(const std::string & nick, const std::string & userhost,
    const std::string & ip) const;
  bool exactMatch(const std::string & nick, const std::string & userhost) const;

  TrapAction getAction() const;
  long getTimeout() const;
  std::string getMask() const;
  std::string getReason() const;
  std::string getString() const;

private:
  TrapAction	action;
  long		timeout;	// For K-Lines only
  std::string	nick;
  std::string	userhost;
  std::string	reason;		// For Kills, K-Lines, and D-Lines only
};


class TrapList
{
public:
  static void cmd(StrList & output, std::string line,
    const std::string & handle = "");
  static bool remove(const std::string & mask);
  static bool remove(const std::string & nick, const std::string & userhost);
  static void clear(void) { TrapList::traps.clear(); };

  static void match(const std::string & nick, const std::string & userhost,
    const std::string & ip);
  static void match(const std::string & nick, const std::string & userhost,
    const BotSock::Address & ip)
  {
    TrapList::match(nick, userhost, BotSock::inet_ntoa(ip));
  };

  static void list(StrList & output);

  static void save(ofstream & file);

  static TrapAction actionType(const std::string & text);
  static std::string actionString(const TrapAction & action);

private:
  static std::list<Trap> traps;

  static Trap add(const TrapAction action, const long timeout,
    const std::string & mask, const std::string & reason);

  static void split(const std::string & mask, std::string & nick,
    std::string & userhost);
};

#endif /* __TRAP_H__ */

