// $Id: userhash.cc,v 1.4 2003/02/23 23:35:52 toast Exp $

// Std C++ Headers
#include <string>
#include <list>
#include <vector>
#include <map>

// Std C Headers
#include <stdio.h>
#include <time.h>

// OOMon Headers
#include "strtype"
#include "userhash.h"
#include "botsock.h"
#include "util.h"
#include "config.h"
#include "vars.h"
#include "autoaction.h"
#include "seedrand.h"
#include "trap.h"
#include "main.h"
#include "log.h"
#include "engine.h"


#define CLONECONNECTCOUNT 3
#define CLONECONNECTFREQ  30

#define CLONEDETECTINC 15


UserHash users;


int
UserHash::hashFunc(const std::string & key)
{
  std::string::size_type len = key.length();
  int i = 0;

  if (len > 0)
  {
    i = key[0];

    if (len > 1)
    {
      i |= (key[1] << 8);

      if (len > 2)
      {
        i |= (key[2] << 16);

	if (len > 3)
	{
          i |= (key[3] << 24);
	}
      }
    }
  }
  return i % HASHTABLESIZE;
}


int
UserHash::hashFunc(const BotSock::Address & key)
{
  return (key & BotSock::vhost_netmask) % HASHTABLESIZE;
}


void
UserHash::add(const std::string & nick, const std::string & userhost,
  const std::string & ip, bool fromTrace, bool isOper,
  const std::string & userClass)
{
  // Find '@' in user@host
  std::string::size_type at = userhost.find('@');

  if (std::string::npos != at)
  {
    // Get username and hostname from user@host
    std::string user = DownCase(userhost.substr(0, at));
    std::string host = DownCase(userhost.substr(at + 1));

    if (!checkForSpoof(nick, user, host, ip))
    {
      // If we made it this far, we'll just assume it isn't a spoofed
      // hostname.  Oh, happy day!

      UserEntry *newuser = new UserEntry;

      if (newuser == (UserEntry *)NULL)
      {
        ::SendAll("Ran out of memory in UserHash::add()", UF_OPER);
        Log::Write("Ran out of memory in UserHash::add()");
        gracefuldie(SIGABRT);
      }

      newuser->nick = nick;
      newuser->user = user;
      newuser->host = host;
      newuser->userClass = DownCase(userClass);
      if (ip != "")
        newuser->ip = BotSock::inet_addr(ip);
      else
        newuser->ip = INADDR_NONE;
      newuser->connectTime = (fromTrace ? 0 : time(NULL));
      newuser->reportTime = 0;
      newuser->linkCount = 0;
      newuser->isOper = isOper;

      int score = seedrandScore(nick);
      newuser->randScore = score;

      // Determine the domain name type
      std::string domain = getDomain(host, false);
      newuser->domain = domain;

      // Add it to the hash tables
      UserHash::addToHash(this->hosttable, host, newuser);
      UserHash::addToHash(this->domaintable, domain, newuser);
      UserHash::addToHash(this->usertable, user, newuser);
      UserHash::addToHash(this->iptable, newuser->ip, newuser);

      // Don't check for wingate or clones when doing a TRACE
      if (!fromTrace)
      {
        if (!Config::IsOKHost(userhost, ip) && !Config::IsOper(userhost, ip))
        {
          // Check if this client matches any traps
          if (vars[VAR_TRAP_CONNECTS]->getBool())
          {
            TrapList::match(nick, userhost, ip);
          }

          if (score >= vars[VAR_SEEDRAND_REPORT_MIN]->getInt())
          {
	    std::string notice("Random (score: " + IntToStr(score) +
	      ") nick connect: " + nick + " (" + userhost + ")");
	    if ("" != ip)
	    {
	      notice = notice + " [" + ip + "]";
	    }
	    ::SendAll(notice, UF_OPER, WATCH_SEEDRAND);
	    Log::Write(notice);

	    doAction(nick, userhost, BotSock::inet_addr(ip),
	      vars[VAR_SEEDRAND_ACTION]->getAction(),
	      vars[VAR_SEEDRAND_ACTION]->getInt(),
	      vars[VAR_SEEDRAND_REASON]->getString(), false);
          }

          if (ip != "")
          {
            CheckProxy(ip, host, nick, userhost);

#ifdef DETECT_VHOST_CLONES
            addIP(ip);
#endif
          }

          // Clonebot check
          this->checkHostClones(host);
        }
      }
    }
  }
}


void
UserHash::updateOper(const std::string & nick, const std::string & userhost,
  bool isOper)
{
  std::string::size_type at = userhost.find('@');

  if (std::string::npos != at)
  {
    std::string user = DownCase(userhost.substr(0, at));
    std::string host = DownCase(userhost.substr(at + 1));

    int index = UserHash::hashFunc(host);

    UserHash::HashRec *find = this->hosttable[index];

    while (NULL != find)
    {
      if ((nick == find->info->nick) && (user == find->info->user) &&
	(host == find->info->host))
      {
        find->info->isOper = true;

        // There shouldn't be any more, so just return
        return;
      }
      find = find->collision;
    }
  }
}


void
UserHash::updateNick(const std::string & oldNick, const std::string & userhost,
  const std::string & newNick)
{
  std::string::size_type at = userhost.find('@');

  if (std::string::npos != at)
  {
    std::string user = DownCase(userhost.substr(0, at));
    std::string host = DownCase(userhost.substr(at + 1));

    int index = UserHash::hashFunc(host);

    UserHash::HashRec *find = this->hosttable[index];

    while (NULL != find)
    {
      if (oldNick == find->info->nick)
      {
        find->info->nick = newNick;

        if (vars[VAR_TRAP_NICK_CHANGES]->getBool())
        {
          TrapList::match(find->info->nick, userhost, find->info->ip);
        }

        int score = seedrandScore(newNick);
        find->info->randScore = score;

        if ((score >= vars[VAR_SEEDRAND_REPORT_MIN]->getInt()) &&
	  !Config::IsOper(userhost) && !Config::IsOKHost(userhost) &&
	  !find->info->isOper)
        {
	  std::string notice("Random (score: " + IntToStr(score) +
	    ") nick change: " + find->info->nick + " (" + userhost + ")");
	  if (INADDR_NONE != find->info->ip)
	  {
	    notice = notice + " [" + BotSock::inet_ntoa(find->info->ip) + "]";
	  }
	  ::SendAll(notice, UF_OPER, WATCH_SEEDRAND);
	  Log::Write(notice);

	  doAction(find->info->nick, userhost, find->info->ip,
	    vars[VAR_SEEDRAND_ACTION]->getAction(),
	    vars[VAR_SEEDRAND_ACTION]->getInt(),
	    vars[VAR_SEEDRAND_REASON]->getString(), false);
        }

        // There shouldn't be any more, so just return
        return;
      }
      find = find->collision;
    }
  }
}


void
UserHash::remove(const std::string & nick, const std::string & userhost,
  const BotSock::Address & ip)
{
  std::string::size_type at = userhost.find('@');

  if (std::string::npos != at)
  {
    std::string user = DownCase(userhost.substr(0, at));
    std::string host = DownCase(userhost.substr(at + 1));
    std::string domain = getDomain(host, false);

    // Set this to true if we are unable to remove the entry from one or
    // more of the tables.
    bool error = false;

    if (!UserHash::removeFromHash(this->hosttable, host, host, user, nick))
    {
      if (!UserHash::removeFromHash(this->hosttable, host, host, user, ""))
      {
        error = true;

#ifdef USERHASH_DEBUG
        cout << "Error removing user from host table: " << nick << " (" <<
	  user << "@" << host << ") [" << BotSock::inet_ntoa(ip) << "]" <<
	  endl;
#endif
      }
    }

    if (!UserHash::removeFromHash(this->domaintable, domain, host, user, nick))
    {
      if (!UserHash::removeFromHash(this->domaintable, domain, host, user, ""))
      {
        error = true;

#ifdef USERHASH_DEBUG
        cout << "Error removing user from domain table: " << nick << " (" <<
	  user << "@" << host << ") [" << BotSock::inet_ntoa(ip) << "]" << endl;
#endif
      }
    }

    if (!UserHash::removeFromHash(this->usertable, user, host, user, nick))
    {
      if (!UserHash::removeFromHash(this->usertable, user, host, user, ""))
      {
	error = true;

#ifdef USERHASH_DEBUG
        cout << "Error removing user from user table: " << nick << " (" <<
	  user << "@" << host << ") [" << BotSock::inet_ntoa(ip) << "]" << endl;
#endif
      }
    }

    if (!UserHash::removeFromHash(this->iptable, ip, host, user, nick))
    {
      if (!UserHash::removeFromHash(this->iptable, ip, host, user, ""))
      {
	error = true;

#ifdef USERHASH_DEBUG
        cout << "Error removing user from IP table: " << nick << " (" <<
	  user << "@" << host << ") [" << BotSock::inet_ntoa(ip) << "]" << endl;
#endif
      }
    }

    // Any errors occur?
    if (error)
    {
      Log::Write("Error removing user from table(s): " + nick + " (" + user +
	"@" + host + ") [" + BotSock::inet_ntoa(ip) + "]");
    }
  }
}


void
UserHash::addToHash(HashRec *table[], const std::string & key, UserEntry *item)
{
  int index = UserHash::hashFunc(key);

  UserHash::HashRec *newhashrec = new UserHash::HashRec;

  if (newhashrec == (HashRec *)NULL)
  {
    ::SendAll("Ran out of memory in UserHash::addToHash()", UF_OPER);
    Log::Write("Ran out of memory in UserHash::addToHash()");
    gracefuldie(SIGABRT);
  }

  newhashrec->info = item;
  if (NULL != newhashrec->info)
  {
    newhashrec->info->linkCount++;
  }
  newhashrec->collision = table[index];
  table[index] = newhashrec;
}


void
UserHash::addToHash(HashRec *table[], const BotSock::Address & key,
  UserEntry *item)
{
  int index = UserHash::hashFunc(key);

  UserHash::HashRec *newhashrec = new UserHash::HashRec;

  if (newhashrec == (HashRec *)NULL)
  {
    ::SendAll("Ran out of memory in UserHash::addToHash()", UF_OPER);
    Log::Write("Ran out of memory in UserHash::addToHash()");
    gracefuldie(SIGABRT);
  }

  newhashrec->info = item;
  if (NULL != newhashrec->info)
  {
    newhashrec->info->linkCount++;
  }
  newhashrec->collision = table[index];
  table[index] = newhashrec;
}


bool
UserHash::removeFromHash(HashRec *table[], const std::string & key,
  const std::string & host, const std::string & user, const std::string & nick)
{
  int index = UserHash::hashFunc(key);

  UserHash::HashRec *find = table[index];

  UserHash::HashRec *prev = NULL;

  while (find)
  {
    if (((host == "") || (find->info->host == host)) &&
      ((user == "") || (find->info->user == user)) &&
      ((nick == "") || (find->info->nick == nick)))
    {
      if (prev)
	prev->collision = find->collision;
      else
	table[index] = find->collision;

      if (find->info->linkCount > 0)
	find->info->linkCount--;

      if (find->info->linkCount == 0)
      {
	delete find->info;
      }
      delete find;
      return true;
    }
    prev = find;
    find = find->collision;
  }
  return false;
}


bool
UserHash::removeFromHash(HashRec *table[], const BotSock::Address & key,
  const std::string & host, const std::string & user, const std::string & nick)
{
  int index = UserHash::hashFunc(key);

  UserHash::HashRec *find = table[index];

  UserHash::HashRec *prev = NULL;

  while (find)
  {
    if (((host == "") || (find->info->host == host)) &&
      ((user == "") || (find->info->user == user)) &&
      ((nick == "") || (find->info->nick == nick)))
    {
      if (prev)
	prev->collision = find->collision;
      else
	table[index] = find->collision;

      if (find->info->linkCount > 0)
	find->info->linkCount--;

      if (find->info->linkCount == 0) {
	delete find->info;
      }
      delete find;
      return true;
    }
    prev = find;
    find = find->collision;
  }
  // Didn't find it, try it without an IP
  if (INADDR_NONE != key)
  {
    removeFromHash(table, INADDR_NONE, host, user, nick);
  }
  return false;
}


void
UserHash::clearHash(HashRec *table[])
{
  for (int i = 0; i < HASHTABLESIZE; ++i)
  {
    HashRec *hp = table[i];
    while (hp)
    {
      HashRec *next_hp = hp->collision;

      if (hp->info->linkCount > 0)
      {
        hp->info->linkCount--;
      }

      if(hp->info->linkCount == 0)
      {
        delete hp->info;
      }
  
      delete hp;
      hp = next_hp;
    } 
    table[i] = NULL;
  }
}


void
UserHash::clear()
{
  UserHash::clearHash(this->usertable);
  UserHash::clearHash(this->hosttable);
  UserHash::clearHash(this->domaintable);
  UserHash::clearHash(this->iptable);
}


int
UserHash::listUsers(StrList & output, const std::string & userhost,
  std::string className, bool kill, const std::string & from,
  const std::string & reason) const
{
  int numfound = 0;

  if ((userhost == "*") || (userhost == "*@*"))
  {
    output.push_back("Listing all users is not recommended. To do it anyway,");
    output.push_back("use 'list ?*@*'.");
  }
  else
  {
    std::string outmsg;

    if (kill)
    {
      Log::Write("KILLLIST " + userhost + " (" + reason + ") [" + from + "]");
    }

    // Convert our class name to uppercase now
    className = DownCase(className);

    for (int index = 0; index < HASHTABLESIZE; ++index)
    {
      UserHash::HashRec *userptr = this->domaintable[index];

      while (userptr)
      {
	std::string fullUH = std::string(userptr->info->user) + "@" +
	  std::string(userptr->info->host);
	std::string IP;
	bool matchesUH = MatchesMask(fullUH, userhost);
	bool matchesUIP = false;
	if (INADDR_NONE != userptr->info->ip)
	{
	  IP = BotSock::inet_ntoa(userptr->info->ip);
	  std::string fullUIP(userptr->info->user + "@" + IP);
	  matchesUIP = MatchesMask(fullUIP, userhost);
	}
	if ((matchesUH || matchesUIP) && ((0 == className.length()) ||
	  (className == userptr->info->userClass)))
	{
	  if (!numfound++)
	  {
	    output.push_back("The following clients match " + userhost + ":");
	  }
	  if (INADDR_NONE != userptr->info->ip)
	  {
	    struct in_addr in;
	    in.s_addr = userptr->info->ip;
	    outmsg = std::string("  ") + userptr->info->nick + " (" + fullUH +
	      ") [" + IP + "]";
	  }
	  else
	  {
	    outmsg = std::string("  ") + userptr->info->nick + " (" + fullUH +
	      ")";
	  }
	  if (kill)
	  {
	    server.kill(from, userptr->info->nick, reason);
	  }
	  output.push_back(outmsg);
	}
	userptr = userptr->collision;
      }
    }
    if (numfound > 0)
    {
      outmsg = IntToStr(numfound) + " matches for " + userhost + " found.";
    }
    else
    {
      outmsg = "No matches for " + userhost + " found.";
    }
    output.push_back(outmsg);
  }

  return numfound;
}


int
UserHash::listNicks(StrList & output, const std::string & nick,
  std::string className) const
{
  std::string outmsg;
  int numfound = 0;

  // Convert class name to uppercase now
  className = DownCase(className);

  for (int index = 0; index < HASHTABLESIZE; ++index)
  {
    UserHash::HashRec *userptr = this->domaintable[index];

    while (userptr)
    {
      if (MatchesMask(userptr->info->nick, nick, true) &&
	((0 == className.length()) || (className == userptr->info->userClass)))
      {
        if (!numfound++)
        { 
          output.push_back("The following clients match " + nick + ":");
        }
	if (INADDR_NONE != userptr->info->ip)
	{
	  outmsg = std::string("  ") + userptr->info->nick + " (" +
	    userptr->info->user + "@" + userptr->info->host + ") [" +
	    BotSock::inet_ntoa(userptr->info->ip) + "]";
	}
	else
	{
	  outmsg = std::string("  ") + userptr->info->nick + " (" +
	    userptr->info->user + "@" + userptr->info->host + ")";
	}
        output.push_back(outmsg);
      }
      userptr = userptr->collision;
    }
  } 
    
  if (numfound > 0)
  {
    outmsg = IntToStr(numfound) + " matches for " + nick + " found.";
  }
  else
  {
    outmsg = "No matches for " + nick + " found.";
  }
  output.push_back(outmsg);

  return numfound;
}


bool
UserHash::have(std::string nick) const
{
  nick = DownCase(nick);

  for (int i = 0; i < HASHTABLESIZE; ++i)
  {
    UserHash::HashRec *find = this->hosttable[i];

    while (find)
    {
      if (DownCase(find->info->nick) == DownCase(nick))
      {
        return true;
      }
      find = find->collision;
    }
  }
  return false;
}


void
UserHash::reportClasses(StrList & output, const std::string & className)
{
  typedef std::map<std::string,int> ClassType;
  ClassType Classes;

  for (int i = 0; i < HASHTABLESIZE; ++i)
  {
    UserHash::HashRec *userptr = this->domaintable[i];
    while (userptr)
    {
      Classes[DownCase(userptr->info->userClass)]++;
      userptr = userptr->collision;
    }
  }

  char outmsg[MAX_BUFF];
  snprintf(outmsg, sizeof(outmsg), "%-10s %-6s %s", "Class", "Count",
    "Description");
  output.push_back(outmsg);

  for (ClassType::iterator pos = Classes.begin(); pos != Classes.end();
    ++pos)
  {
    if ((0 == className.length()) || (pos->first == DownCase(className)))
    {
      snprintf(outmsg, sizeof(outmsg), "%-10s %-6d %s", pos->first.c_str(),
	pos->second, Config::GetYLineDescription(pos->first).c_str());
      output.push_back(outmsg);
    }
  }
}


void
UserHash::reportSeedrand(StrList & output, const std::string & mask,
  int threshold)
{
  output.push_back("Searching " + mask + ".  Threshold is " +
    IntToStr(threshold));

  std::list<UserHash::ScoreNode> scores;

  for (int temp = 0; temp < HASHTABLESIZE; ++temp)
  {
    for (UserHash::HashRec *find = this->usertable[temp]; find;
      find = find->collision)
    {
      if (MatchesMask(find->info->nick, mask, true))
      {
	if (find->info->randScore >= threshold)
        {
	  UserHash::ScoreNode tmp(find->info, find->info->randScore);
	  scores.push_back(tmp);
        }
      }
    }
  }

  scores.sort();

  for (std::list<UserHash::ScoreNode>::iterator pos = scores.begin();
    pos != scores.end(); ++pos)
  {
    UserEntry *info = pos->getInfo();

    output.push_back("  " + IntToStr(pos->getScore(), 4) + " -- " +
      info->nick + " (" + info->user + "@" + info->host + ") [" +
      BotSock::inet_ntoa(info->ip) + "]");
  }

  if (0 == scores.size())
  {
    output.push_back("No matches for " + mask + " found.");
  }
  else if (1 == scores.size())
  {
    output.push_back("1 match for " + mask + " found.");
  }
  else
  {
    output.push_back(IntToStr(scores.size()) + " matches for " + mask +
      " found.");
  }
}


void
UserHash::reportDomains(StrList & output, const int num)
{
  std::list<UserHash::SortEntry> sort;
  int maxCount = 1;
  int foundany = 0;

  if (num < 1)
  {
    output.push_back("*** Parameter must be greater than 0!");
    return;
  }

  for (int i = 0; i < HASHTABLESIZE; ++i)
  {
    HashRec *userptr = this->hosttable[i];

    while (NULL != userptr)
    {
      bool found = false;

      for (std::list<UserHash::SortEntry>::iterator pos = sort.begin();
	pos != sort.end(); ++pos)
      {
        if (userptr->info->domain == pos->domainrec->domain)
	{
	  found = true;
          if (++(pos->count) > maxCount)
	  {
	    maxCount = pos->count;
	  }
	  break;
	}
      }
      if (!found)
      {
	UserHash::SortEntry newEntry;

        newEntry.domainrec = userptr->info;
        newEntry.count = 1;

	sort.push_back(newEntry);
      }

      userptr = userptr->collision;
    }
  }

  /* Print 'em out from highest to lowest */
  while (maxCount >= num)
  {
    int currentCount = maxCount;

    maxCount = 0;

    for (std::list<UserHash::SortEntry>::iterator pos = sort.begin();
      pos != sort.end(); ++pos)
    {
      if (pos->count >= currentCount)
      {
        if (!foundany++)
          output.push_back("Domains with most users on the server:");

        char outmsg[MAX_BUFF];            
        sprintf(outmsg, "  %-40s %3d user%s", pos->domainrec->domain.c_str(),
	  pos->count, (pos->count == 1) ? "" : "s");

        output.push_back(outmsg);

	sort.erase(pos);
	--pos;
      }
      else if (pos->count >= maxCount)
      {
	maxCount = pos->count;
      }
    }
  }

  if (!foundany)
  {  
    char outmsg[MAX_BUFF];            
    sprintf(outmsg, "No domains have %d or more users.", num);
    output.push_back(outmsg);
  }
}


void
UserHash::reportClones(StrList & output)
{
  HashRec *userptr, *top, *temp;
  int numfound, i, j, k;
  char notip, foundany = 0;
  char outmsg[MAX_BUFF];
  std::vector<time_t> connfromhost;

  for (i = 0; i < HASHTABLESIZE; ++i)
  {
    HashRec *top = this->hosttable[i];
    HashRec *userptr = this->hosttable[i];

    while (userptr)
    {
      numfound = 0;
      temp = top;
      while (temp != userptr)
      {
        if ((temp->info->host == userptr->info->host))
	{
	  break;
	}
        else
	{
	  temp = temp->collision;
	}
      }
      if (temp == userptr)
      {
	++numfound;
        connfromhost.push_back(temp->info->connectTime);
        temp = temp->collision;
        while (temp)
	{
          if (temp->info->host == userptr->info->host)
	  {
	    ++numfound;
            connfromhost.push_back(temp->info->connectTime);
	  }
          temp = temp->collision;
        }
        if (numfound > 2)
	{
          for (k = numfound - 1; k > 1; --k)
	  {
            for (j = 0; j < numfound - k; ++j)
	    {
              if (connfromhost[j] &&
                  connfromhost[j] - connfromhost[j+k] <= (k+1) * CLONEDETECTINC)
	      {
                goto getout;  /* goto rules! */
	      }
            }
	  }
          getout:
          if (k > 1)
	  {
            if (!foundany++)
	    {
              output.push_back("Possible clonebots from the following hosts:");
	    }
            sprintf(outmsg,
	      "  %2d connections in %3d seconds (%2d total) from %s", k + 1,
	      connfromhost[j] - connfromhost[j+k], numfound+1,
	      userptr->info->host.c_str());
            output.push_back(outmsg);
          }   
        } 
      }     
      userptr = userptr->collision;
    }
  }
  if (!foundany)
  {
    output.push_back("No potential clonebots found.");
  }
}


void
UserHash::reportMulti(StrList & output, const int minimum)
{
  HashRec *userptr,*top,*temp;
  int numfound,i;
  char notip, foundany = 0;
  char outmsg[MAX_BUFF];
  int minclones = vars[VAR_MULTI_MIN]->getInt();

  if (minimum > 0)
    minclones = minimum;

  for (i=0;i<HASHTABLESIZE;++i) {
    top = userptr = this->domaintable[i];
    while (userptr)
    {
      numfound = 0;
      /* Ensure we haven't already checked this user & domain */
      temp = top;
      while (temp != userptr)
      {
        if ((temp->info->user == userptr->info->user) &&
          (temp->info->domain == userptr->info->domain))
	{
          break;
	}
        else
	{
          temp = temp->collision;
	}
      }
      if (temp == userptr)
      {
        temp = temp->collision;
        while (temp)
	{
          if ((temp->info->user == userptr->info->user) &&
            (temp->info->domain == userptr->info->domain))
	  {
            if (vars[VAR_OPER_IN_MULTI]->getBool() || (!temp->info->isOper &&
	      !Config::IsOper(std::string(temp->info->user) + "@" +
	      std::string(temp->info->host))))
            {
              numfound++;       /* - zaph & Dianora :-) */
	    }
	  }
          temp = temp->collision;
        }
        if (numfound >= (minclones - 1))
	{
          if (!foundany++)
            output.push_back("Multiple clients from the following userhosts:");
          notip = strncmp(userptr->info->domain.c_str(),
	    userptr->info->host.c_str(),
            strlen(userptr->info->domain.c_str())) ||
            (strlen(userptr->info->domain.c_str()) ==
	    strlen(userptr->info->host.c_str()));
          numfound++; /* - zaph and next line*/
          sprintf(outmsg, " %s %2d -- %s@%s%s",
            (numfound > minclones) ? "==>" : "   ", numfound,
	    userptr->info->user.c_str(),
	    notip ? "*" : userptr->info->domain.c_str(),
	    notip ? userptr->info->domain.c_str() : "*", numfound);
          output.push_back(outmsg);
        }
      }  
      userptr = userptr->collision;
    }
  }         
  if (!foundany) 
    output.push_back("No multiple logins found.");
}

void
UserHash::reportHMulti(StrList & output, const int minimum)
{
  HashRec *userptr,*top,*temp;
  int numfound,i;
  char foundany = 0;
  char outmsg[MAX_BUFF];
  int minclones = vars[VAR_MULTI_MIN]->getInt();

  if (minimum > 0)
    minclones = minimum;

  for (i=0;i<HASHTABLESIZE;++i) {
    top = userptr = this->domaintable[i];
    while (userptr)
    {
      numfound = 0;
      // Ensure we haven't already checked this hostname
      temp = top;
      while (temp != userptr)
        if ((temp->info->host == userptr->info->host))
          break;
        else
          temp = temp->collision;
      if (temp == userptr) {
        temp = temp->collision;
        while (temp) {
          if ((temp->info->host == userptr->info->host))
	  {
            if (vars[VAR_OPER_IN_MULTI]->getBool() || (!temp->info->isOper &&
	      !Config::IsOper(std::string(temp->info->user) + "@" +
	      std::string(temp->info->host))))
            {
              numfound++;       /* - zaph & Dianora :-) */
	    }
	  }
          temp = temp->collision;
        }
        if (numfound >= (minclones - 1))
	{
          if (!foundany++)
            output.push_back("Multiple clients from the following hosts:");
          numfound++; /* - zaph and next line*/
          sprintf(outmsg, " %s %2d -- *@%s",
	    (numfound > minclones) ? "==>" : "   ", numfound,
	    userptr->info->host.c_str());
          output.push_back(outmsg);
        }
      }  
      userptr = userptr->collision;
    }
  }         
  if (!foundany) 
    output.push_back("No multiple logins found.");
}

void
UserHash::reportUMulti(StrList & output, const int minimum)
{
  HashRec *userptr,*top,*temp;
  int numfound,i;
  char foundany = 0;
  char outmsg[MAX_BUFF];
  int minclones = vars[VAR_MULTI_MIN]->getInt();

  if (minimum > 0)
    minclones = minimum;

  for (i=0;i<HASHTABLESIZE;++i) {
    top = userptr = this->usertable[i];
    while (userptr)
    {
      numfound = 0;
      // Ensure we haven't already checked this username
      temp = top;
      while (temp != userptr)
        if ((temp->info->user == userptr->info->user))
          break;
        else
          temp = temp->collision;
      if (temp == userptr) {
        temp = temp->collision;
        while (temp) {
          if ((temp->info->user == userptr->info->user))
	  {
            if (vars[VAR_OPER_IN_MULTI]->getBool() || (!temp->info->isOper &&
	      !Config::IsOper(std::string(temp->info->user) + "@" +
	      std::string(temp->info->host))))
            {
              numfound++;       /* - zaph & Dianora :-) */
	    }
	  }
          temp = temp->collision;
        }
        if (numfound >= (minclones - 1))
	{
          if (!foundany++)
            output.push_back("Multiple clients from the following usernames:");
          numfound++; /* - zaph and next line*/
          sprintf(outmsg, " %s %2d -- %s@*",
	    (numfound > minclones) ? "==>" : "   ", numfound,
	    userptr->info->user.c_str());
          output.push_back(outmsg);
        }
      }  
      userptr = userptr->collision;
    }
  }         
  if (!foundany) 
    output.push_back("No multiple logins found.");
}

void
UserHash::reportVMulti(StrList & output, const int minimum)
{
  HashRec *userptr,*top,*temp;
  int numfound,i;
  char notip, foundany = 0;
  char outmsg[MAX_BUFF];
  int minclones = vars[VAR_MULTI_MIN]->getInt();
  struct in_addr in;

  if (minimum > 0)
    minclones = minimum;

  for (i=0;i<HASHTABLESIZE;++i) {
    top = userptr = this->iptable[i];
    while (userptr)
    {
      numfound = 0;
      /* Ensure we haven't already checked this user & domain */
      temp = top;
      while (temp != userptr)
        if ((temp->info->user == userptr->info->user) &&
          ((userptr->info->ip & BotSock::vhost_netmask) ==
	  (userptr->info->ip & BotSock::vhost_netmask)))
	{
          break;
	}
        else
	{
          temp = temp->collision;
	}
      if (temp == userptr) {
        temp = temp->collision;
        while (temp) {
          if ((temp->info->user == userptr->info->user) &&
            ((userptr->info->ip & BotSock::vhost_netmask) ==
	    (userptr->info->ip & BotSock::vhost_netmask)))
	  {
            if (vars[VAR_OPER_IN_MULTI]->getBool() || (!temp->info->isOper &&
	      !Config::IsOper(std::string(temp->info->user) + "@" +
	      std::string(temp->info->host))))
            {
              numfound++;       /* - zaph & Dianora :-) */
	    }
	  }
          temp = temp->collision;
        }
        if (numfound >= (minclones - 1))
	{
          if (!foundany++)
            output.push_back("Multiple clients from the following vhosts:");
          numfound++; /* - zaph and next line*/
	  std::string IP = BotSock::inet_ntoa(userptr->info->ip);
	  std::string::size_type lastDot = IP.rfind('.');
	  if (std::string::npos != lastDot)
	  {
	    IP = IP.substr(0, lastDot);
	  }
          sprintf(outmsg, " %s %2d -- %s@%s.*",
	    (numfound > minclones) ? "==>" : "   ", numfound,
	    userptr->info->user.c_str(), IP.c_str());
          output.push_back(outmsg);
        }
      }  
      userptr = userptr->collision;
    }
  }         
  if (!foundany) 
    output.push_back("No multiple logins found.");
}


void
UserHash::checkHostClones(const std::string & host)
{
  UserHash::HashRec *find;
  int clonecount = 0; 
  int reportedclones = 0;
  std::string lastUser;
  bool last_identd,current_identd;
  bool different;
  time_t now, lastreport, oldest;
  struct tm *tmrec;
  int index;
  
  oldest = now = time(NULL);
  lastreport = 0;
  index = UserHash::hashFunc(host);
  find = this->hosttable[index];
  
  while (NULL != find)
  {
    if ((find->info->host == host) &&
      (now - find->info->connectTime < CLONECONNECTFREQ + 1))
    {
      if (find->info->reportTime > 0)
      {
        ++reportedclones;
        if (lastreport < find->info->reportTime)
          lastreport = find->info->reportTime;
      }
      else
      {
        ++clonecount;
        if (find->info->connectTime < oldest)
        {
          oldest = find->info->connectTime;
        }
      }
    }
    find = find->collision;
  }
  
  if ((reportedclones == 0 && clonecount < CLONECONNECTCOUNT) ||
    now - lastreport < 10)
  {
    return;
  }

  std::string notice;
  if (reportedclones)
  {
    notice = IntToStr(clonecount) + " more possible clones (" +
      IntToStr(clonecount + reportedclones) + " total) from " + host + ":";
  }
  else
  {
    notice = "Possible clones from " + host + " detected: " +
      IntToStr(clonecount) + " connects in " + IntToStr(now - oldest) +
      " seconds";
  }
  ::SendAll(notice, UF_OPER);
  Log::Write(notice);
            
  clonecount = 0;
  find = this->hosttable[index];
    
  while (find)
  {
    if ((find->info->host == host) &&
      (now - find->info->connectTime < CLONECONNECTFREQ + 1) &&
      find->info->reportTime == 0)
    {
      ++clonecount;
      tmrec = localtime(&find->info->connectTime);
          
      char notice[MAX_BUFF];
      char notice1[MAX_BUFF];
      if(clonecount == 1)  
      { 
        sprintf(notice1,"  %s is %s@%s (%2.2d:%2.2d:%2.2d)",
          find->info->nick.c_str(), find->info->user.c_str(),
          find->info->host.c_str(), tmrec->tm_hour, tmrec->tm_min,
	  tmrec->tm_sec);
      }
      else
      {
        sprintf(notice,"  %s is %s@%s (%2.2d:%2.2d:%2.2d)",
	  find->info->nick.c_str(), find->info->user.c_str(),
	  find->info->host.c_str(), tmrec->tm_hour, tmrec->tm_min,
	  tmrec->tm_sec);
      }
  
      last_identd = current_identd = true;
      different = false;
        
      if (1 == clonecount)
      {
        lastUser = find->info->user;
      }
      else if (2 == clonecount)
      {
        std::string currentUser;

        if (lastUser[0] == '~')
        {
          lastUser = lastUser.substr(1);
          last_identd = false;
        }

        currentUser = find->info->user;
        if (currentUser[0] == '~')
        {
          currentUser = currentUser.substr(1);
          current_identd = false;
        }

        if (lastUser != currentUser)
        {
          different = true;
        }

        suggestKline(true, find->info->user, find->info->host, different,
	  last_identd | current_identd, MK_CLONES);
      }

      find->info->reportTime = now;
      if(clonecount == 1)
      {
	// do nothing
      }
      else if(clonecount == 2) 
      {   
        ::SendAll(notice1, UF_OPER);
        ::SendAll(notice, UF_OPER);
        Log::Write(notice1);
        Log::Write(notice);
      }
      else if (clonecount < 5)
      {
        ::SendAll(notice, UF_OPER);
        Log::Write(notice);
      }
      else if (clonecount == 5)
      {
        ::SendAll(notice, UF_OPER);
        Log::Write(notice);
      }
    }
    find = find->collision;
  }
}


void
UserHash::status(StrList & output)
{
  int userCount = 0;
  long scoreSum = 0;
  for (int index = 0; index < HASHTABLESIZE; ++index)
  {
    for (UserHash::HashRec *pos = this->usertable[index]; NULL != pos;
      pos = pos->collision)
    {
      scoreSum += pos->info->randScore;
      userCount++;
    }
  }
  output.push_back("Users in usertable: " + IntToStr(userCount));
  int totalUsers = userCount;

  userCount = 0;
  for (int index = 0; index < HASHTABLESIZE; ++index)
  {
    for (UserHash::HashRec *pos = this->hosttable[index]; NULL != pos;
      pos = pos->collision)
    {
      userCount++;
    }
  }
  output.push_back("Users in hosttable: " + IntToStr(userCount));

  userCount = 0;
  for (int index = 0; index < HASHTABLESIZE; ++index)
  {
    for (UserHash::HashRec *pos = this->domaintable[index]; NULL != pos;
      pos = pos->collision)
    {
      userCount++;
    }
  }
  output.push_back("Users in domaintable: " + IntToStr(userCount));

  userCount = 0;
  for (int index = 0; index < HASHTABLESIZE; ++index)
  {
    for (UserHash::HashRec *pos = this->iptable[index]; NULL != pos;
      pos = pos->collision)
    {
      userCount++;
    }
  }
  output.push_back("Users in iptable: " + IntToStr(userCount));

  if (totalUsers > 0)
  {
    output.push_back("Average seedrand score: " +
      IntToStr(scoreSum / totalUsers));
  }
  else
  {
    output.push_back("Average seedrand score: N/A");
  }
}


BotSock::Address
UserHash::getIP(std::string nick, const std::string & userhost) const
{
  std::string::size_type at = userhost.find('@');

  if (std::string::npos != at)
  {
    std::string user = DownCase(userhost.substr(0, at));
    std::string host = DownCase(userhost.substr(at + 1));

    nick = DownCase(nick);

    int index = UserHash::hashFunc(host);

    UserHash::HashRec *find = this->hosttable[index];

    while (NULL != find)
    {
      if ((nick == DownCase(find->info->nick)) && (user == find->info->user) &&
	(host == find->info->host))
      {
        // Found the user -- return its IP address
        return find->info->ip;
      }
      find = find->collision;
    }
  }
  return INADDR_NONE;
}

