// $Id: userhash.h,v 1.3 2003/02/17 17:22:57 toast Exp $

#ifndef __USERHASH_H__
#define __USERHASH_H__

// Std C++ Headers
#include <string>

// Std C Headers
#include <time.h>

// OOMon Headers
#include "strtype"
#include "botsock.h"


#define HASHTABLESIZE 3001


class UserHash
{
public:
  UserHash() { };
  ~UserHash() { this->clear(); };

  void clear();

  void add(const std::string & nick, const std::string & userhost,
    const std::string & ip, bool fromTrace, bool isOper,
    const std::string & userClass);

  BotSock::Address getIP(std::string nick, const std::string & userhost) const;

  void updateNick(const std::string & oldNick, const std::string & userhost,
    const std::string & newNick);
  void updateOper(const std::string & nick, const std::string & userhost,
    bool isOper);

  void remove(const std::string & nick, const std::string & userhost,
    const BotSock::Address & ip);

  bool have(std::string nick) const;

  void checkHostClones(const std::string & host);

  int listUsers(StrList & output, const std::string & userhost,
    std::string className, bool kill = false, const std::string & from = "",
    const std::string & reason = "") const;
  int listNicks(StrList & output, const std::string & nick,
    std::string className) const;

  void reportClasses(StrList & output, const std::string & className);
  void reportSeedrand(StrList & output, const std::string & mask,
    int threshold);
  void reportDomains(StrList & output, const int num);
  void reportClones(StrList & output);
  void reportMulti(StrList & output, const int minimum);
  void reportUMulti(StrList & output, const int minimum);
  void reportHMulti(StrList & output, const int minimum);
  void reportVMulti(StrList & output, const int minimum);

  void status(StrList & output);

private:
  class UserEntry
  {
  public:
    std::string nick;
    std::string user;
    std::string host;
    std::string domain;
    std::string userClass;
    BotSock::Address ip;
    time_t connectTime;
    time_t reportTime;
    bool isOper;
    int randScore;
    int linkCount;
  };

  class HashRec
  {
  public:
    UserEntry *info;
    HashRec *collision;
  };

  class ScoreNode
  {
  public:
    ScoreNode(UserEntry *info, int score)
    {
      this->info = info;
      this->score = score;
    };
    UserEntry *getInfo() const { return this->info; };
    int getScore() const { return this->score; };
    bool operator < (const ScoreNode & compare) const
      { return (this->score < compare.score); };
  private:
    UserEntry *info;
    int score;
  };

  class SortEntry
  {
  public:
    UserEntry *domainrec;
    int count;
  };

  static int hashFunc(const BotSock::Address & key);
  static int hashFunc(const std::string & key);

  static void addToHash(HashRec *table[], const BotSock::Address & key,
    UserEntry *item);
  static void addToHash(HashRec *table[], const std::string & key,
    UserEntry *item);

  static bool removeFromHash(HashRec *table[], const std::string & key,
    const std::string & host, const std::string & user,
    const std::string & nick);
  static bool removeFromHash(HashRec *table[], const BotSock::Address & key,
    const std::string & host, const std::string & user,
    const std::string & nick);

  static void clearHash(HashRec *table[]);

  HashRec *hosttable[HASHTABLESIZE];
  HashRec *domaintable[HASHTABLESIZE];
  HashRec *usertable[HASHTABLESIZE];
  HashRec *iptable[HASHTABLESIZE];
};


extern UserHash users;


#endif /* __USERHASH_H__ */

