// $Id: util.cc,v 1.50 2003/05/10 21:01:50 toast Exp $

// C headers
#include <sys/time.h>
#include <unistd.h>
#include <time.h>

// C++ headers
#include <string>

#include "strtype"
#include "oomon.h"
#include "util.h"
#include "remote.h"

// Originally I used strsep() to do this, but quickly learned how
// unportable it makes my code. *sigh* oh well. This works too.
int
StrSplit(StrVector & temp, std::string input, const std::string & tokens,
  bool ignoreEmpties)
{
  int count = 0;

  temp.clear();

  while (input != "")
  {
    std::string::size_type pos = input.find_first_of(tokens);

    if (std::string::npos == pos)
    {
      temp.push_back(input);
      input = "";
      count++;
    }
    else
    {
      std::string unit = input.substr(0, pos);
      std::string rest = input.substr(pos + 1);

      if (!ignoreEmpties || (unit != ""))
      {
        temp.push_back(unit);
        count++;
      }
      input = rest;
    }
  }
  return count;
}


int
ReadLine(register int fd, std::string & text)
{
  int n, rc;
  char c;

  text = "";

  for (n = 1; n < MAX_BUFF; n++) {
    if ((rc = read(fd, &c, 1)) == 1) {
      if ((c != '\n') && (c != '\r'))
	text = text + c;
      else {
	//n--;
	break;
      }
    } else if (rc == 0) {
      if (n == 1)
	return 0;
      else
	break;
    } else
      return -1;
  }
  return n;
}


int
SplitIRC(StrVector & temp, std::string text)
{
  int count = 0;

  temp.clear();
  while (text.length() > 0)
  {
    if (text.at(0) == ':')
    {
      text.erase((std::string::size_type) 0, 1);
      temp.push_back(text);
      text = std::string("");
      count++;
    }
    else
    {
      std::string::size_type i = text.find(' ');
      if (std::string::npos != i)
      {
	if (i == 0)
	{
	  text.erase((std::string::size_type) 0, 1);
	}
	else
	{
	  std::string chunk = text.substr(0, i);
          temp.push_back(chunk);
          text.erase(0, i + 1);
          count++;
	}
      }
      else
      {
        temp.push_back(text);
        text = "";
        count++;
      }
    }
  }
  return count;
}


void
SplitFrom(std::string input, std::string & Nick, std::string & UserHost)
{
  int i = input.find('!');
  if (std::string::npos == i)
  {
    // This is a server, not a nick!user@host
    Nick = input;
    UserHost = "";
  }
  else
  {
    // This is a nick!user@host
    Nick = input.substr(0, i);
    UserHost = input.substr(i + 1);
  }
}


std::string
getNick(const std::string & text)
{
  std::string::size_type bang = text.find('!');

  if (std::string::npos == bang)
  {
    return text;
  }
  else
  {
    return text.substr(0, bang);
  }
}


int
IRCCommand(const std::string & text)
{
  if (text == std::string("PING"))
    return IRC_PING;
  else if (text == std::string("NICK"))
    return IRC_NICK;
  else if (text == std::string("NOTICE"))
    return IRC_NOTICE;
  else if (text == std::string("PRIVMSG"))
    return IRC_PRIVMSG;
  else if (text == std::string("WALLOPS"))
    return IRC_WALLOPS;
  else
    return IRC_UNKNOWN;
}


unsigned long
atoul(const char *text)
{
  unsigned long temp = 0;
  int len = strlen(text);
  while ((*text >= '0') && (*text <= '9'))
    temp = temp * 10 + (*(text++) - '0');
  return temp;
}


std::string
FirstWord(std::string & Text)
{
  std::string temp;
  int i = Text.find(' ');
  if (std::string::npos != i)
  {
    temp = Text.substr(0, i);
    Text = Text.substr(i + 1);
  }
  else
  {
    temp = Text;
    Text = "";
  }
  while ((Text.length() > 0) && (Text[0] == ' '))
    Text.erase((std::string::size_type) 0, 1);
  return temp;
}


std::string
UpCase(const std::string & Text)
{
  std::string result = Text;
  std::string::size_type len = Text.length();

  for (std::string::size_type index = 0; index < len; index++)
  {
	if ((result[index] >= 'a') && (result[index] <= 'z'))
	{
		result[index] = (result[index] - 'a' + 'A');
	}
  }

  return result;
}


std::string
DownCase(const std::string & Text)
{
  std::string result = Text;
  std::string::size_type len = Text.length();

  for (std::string::size_type index = 0; index < len; index++)
  {
	if ((result[index] >= 'A') && (result[index] <= 'Z'))
	{
		result[index] = (result[index] - 'A' + 'a');
	}
  }

  return result;
}


bool
Same(const std::string & Text1, const std::string & Text2)
{
  return (UpCase(Text1) == UpCase(Text2));
}


bool
Same(const std::string & Text1, const std::string & Text2,
  const std::string::size_type length)
{
  if ((length > Text1.length()) || (length > Text2.length()))
  {
    return false;
  }

  return (UpCase(Text1.substr(0, length)) == UpCase(Text2.substr(0, length)));
}


static bool
matchesChar(char test, char mask)
{
  return ((mask == '?') || (test == mask));
}


static bool
matchesSpecialChar(char test, char mask)
{
  if ((mask == '?') || (test == mask))
  {
    return true;
  }
  else if ((mask == '#') && isdigit(test))
  {
    return true;
  }
  else if ((mask == '&') && isalpha(test))
  {
    return true;
  }
  else if ((mask == '%') && isalnum(test))
  {
    return true;
  }

  return false;
}


static bool
FixedMatch(std::string TEST, std::string MASK, bool special)
{
  if (TEST == MASK)
  {
    return true;
  }
  else
  {
    if (TEST.length() == MASK.length())
    {
      if (special)
      {
        for (int i = 0; i < MASK.length(); i++)
        {
	  if (!matchesSpecialChar(TEST[i], MASK[i]))
	  {
	    return false;
	  }
	}
      }
      else
      {
        for (int i = 0; i < MASK.length(); i++)
        {
	  if (!matchesChar(TEST[i], MASK[i]))
	  {
	    return false;
	  }
	}
      }
    }
    else
    {
      return false;
    }
  }
  return true;
}


static bool
SMatch(std::string TEST, std::string MASK, bool special)
{
  if ((MASK == "*") || (TEST == MASK))
    return true;
  if ((MASK == "") || (TEST == ""))
    return false;
  if ((MASK[0] == '*') && (std::string::npos != MASK.substr(1).find('*')))
  {
    MASK.erase((std::string::size_type) 0, 1);
    for (int i = 0; i < TEST.length(); i++)
    {
      std::string temps = TEST.substr(i);
      int p = MASK.find('*');
      if ((std::string::npos != p) &&
	FixedMatch(temps.substr(0, p), MASK.substr(0, p), special))
      {
	temps = temps.substr(p);
	std::string tempm = MASK.substr(p);
	if (SMatch(temps, tempm, special))
	  return true;
      }
    }
    return false;
  }
  else if (MASK[0] == '*')
  {
    MASK.erase((std::string::size_type) 0, 1);
    TEST.erase(0, TEST.length() - MASK.length());
    return FixedMatch(TEST, MASK, special);
  }
  else
  {
    int p = MASK.find('*');
    if (std::string::npos != p)
    {
      std::string temps = TEST.substr(0, p);
      std::string tempm = MASK.substr(0, p);
      if (FixedMatch(temps, tempm, special))
      {
        TEST.erase(0, p);
        MASK.erase(0, p);
        return SMatch(TEST, MASK, special);
      }
      else
      {
        return false;
      }
    }
    else
    {
      return FixedMatch(TEST, MASK, special);
    }
  }
}


bool
MatchesMask(std::string TEST, std::string MASK, bool special)
{
  return SMatch(UpCase(TEST), UpCase(MASK), special);
}


bool
ChkPass(std::string CORRECT, std::string TEST)
{
  if (CORRECT == "*")
    return false;
  if ((CORRECT == "") && (TEST == ""))
    return true;
#ifdef USE_CRYPT
  if (CORRECT == crypt(TEST.c_str(), CORRECT.c_str()))
    return true;
#else
  return (CORRECT == TEST);
#endif
  return false;
}


std::string
timeStamp(const std::string & format)
{
  char ts_holder[MAX_BUFF];
  struct tm *time_val;
  time_t t;
  
  t = time(NULL);
  time_val = localtime(&t);

  strftime(ts_holder, MAX_BUFF-1, format.c_str(), time_val);

  return std::string(ts_holder);
}


std::string
IntToStr(int i, std::string::size_type len)
{
  if (i < 0)
  {
    return std::string("-") + IntToStr(-i, (len > 1) ? (len - 1) : 0);
  }
  else
  {
    std::string result;

    if (i == 0)
    {
      result = "0";
    }
    else
    {
      while (i > 0)
      {
        result.insert(result.begin(), 1,
	  (std::string::value_type)((i % 10) + '0'));
        i /= 10;
      }
    }

    std::string::size_type l = result.length();
    if (l < len)
    {
      result.insert(result.begin(), len - l, ' ');
    }
    return result;
  }
}


std::string
ULongToStr(unsigned long i, std::string::size_type len)
{
  std::string result;

  if (i == 0)
  {
    result = "0";
  }
  else
  {
    while (i != 0)
    {
      result.insert(result.begin(), 1,
	(std::string::value_type)((i % 10) + '0'));
      i /= 10;
    }
  }

  std::string::size_type l = result.length();
  if (l < len)
  {
    result.insert(result.begin(), len - l, ' ');
  }
  return result;
}


int
RemoteCommand(const std::string & text)
{
  if (text == "ERROR") return RC_ERROR;
  else if (text == REM_STATUS) return RC_STATUS;
  else if (text == REM_CHAT) return RC_CHAT;
  else if (text == REM_WHO) return RC_WHO;
  else if (text == REM_LINKS) return RC_LINKS;
  else if (text == REM_KLINE) return RC_KLINE;
  else if (text == REM_KCLONE) return RC_KCLONE;
  else if (text == REM_KFLOOD) return RC_KFLOOD;
  else if (text == REM_KBOT) return RC_KBOT;
  else if (text == REM_KSPAM) return RC_KSPAM;
  else if (text == REM_KLINK) return RC_KLINK;
  else if (text == REM_KTRACE) return RC_KTRACE;
  else if (text == REM_KMOTD) return RC_KMOTD;
  else if (text == REM_KINFO) return RC_KINFO;
  else if (text == REM_KPROXY) return RC_KPROXY;
  else if (text == REM_KLINE) return RC_KPERM;
  else if (text == REM_UNKLINE) return RC_UNKLINE;
  else if (text == REM_GLINE) return RC_GLINE;
  else if (text == REM_GCLONE) return RC_GCLONE;
  else if (text == REM_GFLOOD) return RC_GFLOOD;
  else if (text == REM_GBOT) return RC_GBOT;
  else if (text == REM_GSPAM) return RC_GSPAM;
  else if (text == REM_GTRACE) return RC_GTRACE;
  else if (text == REM_GMOTD) return RC_GMOTD;
  else if (text == REM_GINFO) return RC_GINFO;
  else if (text == REM_GPROXY) return RC_GPROXY;
  else if (text == REM_GLINK) return RC_GLINK;
  else if (text == REM_GPERM) return RC_GPERM;
  else if (text == REM_UNGLINE) return RC_UNGLINE;
  else if (text == REM_CONNECT) return RC_CONNECT;
  else if (text == REM_DISCONN) return RC_DISCONN;
  else if (text == REM_DIE) return RC_DIE;
  else if (text == REM_MOTD) return RC_MOTD;
  return RC_UNKNOWN;
}


std::string
GetBotHB(std::string text)
{
  if (text == "*")
    return text;

  int i;
  if ((i = text.find('@')) != std::string::npos)
    return text.substr(i + 1);
  else
    return "";
}


std::string
GetHandleHB(std::string text)
{
  if (text == "*")
    return text;

  int i;
  if ((i = text.find('@')) != std::string::npos)
    return text.substr(0, i);
  else
    return "";
}


std::string::size_type
CountChars(const std::string & text, const std::string::value_type ch)
{
  std::string::size_type length = text.length();
  std::string::size_type count = 0;

  for (std::string::size_type index = 0; index < length; index++)
  {
    if (text[index] == ch)
    {
      count++;
    }
  }

  return count;
}


bool
isIP(const std::string & host)
{
  return ((std::string::npos == host.find_first_not_of(".0123456789")) &&
    (3 == CountChars(host, '.')));
}


std::string
getDomain(std::string host, bool withDot)
{
  if (isIP(host))
  {
    // 1.2.3.4
    std::string::size_type lastDot = host.rfind('.');
    if (std::string::npos != lastDot)
    {
      return (host.substr(0, lastDot + (withDot ? 1 : 0)));
    }
    else
    {
      // This should never happen
      return host;
    }
  }
  else
  {
    std::string::size_type lastDot = host.rfind('.');
    if (std::string::npos == lastDot)
    {
      // domain
      return host;
    }
    else
    {
      std::string::size_type prevDot = host.rfind('.', lastDot - 1);
      if (std::string::npos == prevDot)
      {
        // domain.tld
        return host;
      }
      else
      {
        // *.domain.tld
        std::string tld = host.substr(lastDot + 1, std::string::npos);
        if (tld.length() == 2)
        {
          // *.domain.xx
          std::string::size_type triDot = host.rfind('.', prevDot - 1);
          if (std::string::npos == triDot)
          {
            // user@foo.domain.xx
            if (std::string::npos == host.substr(0,
              prevDot).find_first_of("0123456789"))
            {
              // alpha.domain.xx
	      return (host);
	    }
	    else
	    {
              // al55pha.domain.xx
	      return (host.substr(prevDot + (withDot ? 0 : 1),
		std::string::npos));
	    }
	  }
	  else
	  {
	    // *.foo.domain.xx
	    if (std::string::npos == host.substr(triDot + 1,
              prevDot - triDot - 1).find_first_of("0123456789"))
            {
              // *.alpha.domain.xx
	      return (host.substr(triDot + (withDot ? 0 : 1),
		std::string::npos));
	    }
	    else
	    {
              // *.al55pha.domain.xx
	      return (host.substr(prevDot + (withDot ? 0 : 1),
		std::string::npos));
	    }
	  }
	}
	else
	{
	  // *.domain.tld
	  return (host.substr(prevDot + (withDot ? 0 : 1), std::string::npos));
	}
      }
    }
  }
}


std::string klineMask(const std::string & userhost)
{
  std::string::size_type at = userhost.find('@');
  if (at != std::string::npos)
  {
    std::string user = userhost.substr(0, at);
    std::string host = userhost.substr(at + 1, std::string::npos);

    if ((user.length() > 0) && (user[0] == '~'))
    {
      user = "~*";
    }
    else if (user.length() > 8)
    {
      user = "*" + user.substr(user.length() - 8);
    }
    else
    {
      user = "*" + user;
    }

    std::string domain = getDomain(host, true);

    if (isIP(host))
    {
      return (user + "@" + domain + "*");
    }
    else
    {
      return (user + "@*" + domain);
    }
  }
  else
  {
    return userhost;
  }
}

std::string classCMask(const std::string & ip)
{
  std::string::size_type lastDot = ip.rfind('.');

  return (ip.substr(0, lastDot) + ".*");
}

std::string
StrJoin(char token, const StrVector & items)
{
  std::string result;

  for (StrVector::size_type index = 0; index < items.size(); index++)
  {
    if (index == 0)
      result = items[index];
    else
      result += " " + items[index];
  }

  return result;
}


bool
isNumeric(const std::string & text)
{
  for (std::string::size_type pos = 0; pos < text.length(); ++pos)
  {
    if (!isdigit(text[pos]))
    {
      return false;
    }
  }
  return true;
}

