// $Id: util.h,v 1.26 2003/05/10 21:01:50 toast Exp $

#ifndef __UTIL_H__
#define __UTIL_H__

#include <string>

#include "strtype"

// Split tokenized (char *) into a StrVector
int StrSplit(StrVector &, std::string, const std::string &,
  bool ignoreEmpties = false);

// Join StrVector into a std::string
std::string StrJoin(char, const StrVector &);

// Read a line of data from a socket (or whatever)
int ReadLine(register int, std::string &);

// Splits data from an IRC server into its pieces
int SplitIRC(StrVector &, std::string);

// Splits a nick!user@host into a nick and a user@host
void SplitFrom(std::string, std::string &, std::string &);

// Returns the nick in a nick!user@host thingy
std::string getNick(const std::string &);

// Convers an IRC command to an integer
int IRCCommand(const std::string &);

// Converts ASCII to an unsigned long
unsigned long atoul(const char *);

// Retrieves the first word in a string
std::string FirstWord(std::string &);

// Converts a string to uppercase/lowercase
std::string UpCase(const std::string &);
std::string DownCase(const std::string &);

// Compares two strings (case-insensitive)
bool Same(const std::string & Text1, const std::string & Text2);
bool Same(const std::string & Text1, const std::string & Text2,
  const std::string::size_type length);

// Matches a user@host to a mask (case-insensitive)
bool MatchesMask(std::string, std::string, bool special = false);

// Compares two passwords
bool ChkPass(std::string, std::string);

// Returns a National Representation of the date and time
std::string timeStamp(const std::string & format = "%c");

// Returns a string representation of an integer
std::string IntToStr(int i, std::string::size_type len = 0);
std::string ULongToStr(unsigned long, std::string::size_type len = 0);

int RemoteCommand(const std::string &);

std::string GetBotHB(std::string);
std::string GetHandleHB(std::string);

std::string::size_type CountChars(const std::string & text,
  const std::string::value_type ch);

bool isIP(const std::string & host);
std::string getDomain(std::string host, bool withDot = false);
std::string klineMask(const std::string & userhost);
std::string classCMask(const std::string & ip);

bool isNumeric(const std::string & text);


#define IRC_UNKNOWN	0
#define IRC_PING	1
#define IRC_NICK	2
#define IRC_NOTICE	3
#define IRC_PRIVMSG	4
#define IRC_WALLOPS	5

#endif /* __UTIL_H__ */

